# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

try:
    from neon_speech.stt import *
except ImportError:
    from mycroft.stt import *

# class STT(metaclass=ABCMeta):
#     """ STT Base class, all  STT backends derives from this one. """
#     def __init__(self, config=None):
#         config_core = config or get_neon_speech_config()
#         metric_upload = config_core.get("metric_upload", False)
#         if metric_upload:
#             server_addr = config_core.get("remote_server", "64.34.186.120")
#             self.server_bus = MessageBusClient(host=server_addr)
#             self.server_bus.run_in_thread()
#         else:
#             self.server_bus = None
#         self.lang = str(self.init_language(config_core))
#         config_stt = config_core.get("stt", {})
#         module = config_stt.get("module", "")
#         if "google_cloud" in module:
#             module = "google_cloud"
#         self.config = config_stt.get(module, {})
#         self.credential = self.config.get("credential", {})
#         self.recognizer = Recognizer()
#         self.can_stream = False
#         self.keys = config_core.get("keys", {})
#
#     @staticmethod
#     def init_language(config_core):
#         lang = config_core.get("lang", "en-US")
#         langs = lang.split("-")
#         if len(langs) == 2:
#             return langs[0].lower() + "-" + langs[1].upper()
#         return lang
#
#     @abstractmethod
#     def execute(self, audio, language=None):
#         pass
#
#
# class TokenSTT(STT, metaclass=ABCMeta):
#     def __init__(self):
#         super(TokenSTT, self).__init__()
#         self.token = self.credential.get("token")
#
#
# class GoogleJsonSTT(STT, metaclass=ABCMeta):
#     def __init__(self):
#         super(GoogleJsonSTT, self).__init__()
#         if not self.credential.get("json") or self.keys.get("google_cloud"):
#             self.credential["json"] = self.keys["google_cloud"]
#         self.json_credentials = json.dumps(self.credential.get("json"))
#
#
# class BasicSTT(STT, metaclass=ABCMeta):
#
#     def __init__(self):
#         super(BasicSTT, self).__init__()
#         self.username = str(self.credential.get("username"))
#         self.password = str(self.credential.get("password"))
#
#
# class KeySTT(STT, metaclass=ABCMeta):
#
#     def __init__(self):
#         super(KeySTT, self).__init__()
#         self.id = str(self.credential.get("client_id"))
#         self.key = str(self.credential.get("client_key"))
#
#
# class StreamingSTT(STT, metaclass=ABCMeta):
#     """
#         ABC class for threaded streaming STT implemenations.
#     """
#
#     def __init__(self, results_event, config=None):
#         super().__init__(config)
#         self.stream = None
#         self.can_stream = True
#         self.queue = None
#         self.results_event = results_event
#
#     def stream_start(self, language=None):
#         self.stream_stop()
#         self.queue = Queue()
#         self.stream = self.create_streaming_thread()
#         self.stream.start()
#
#     def stream_data(self, data):
#         self.queue.put(data)
#
#     def stream_stop(self):
#         if self.stream is not None:
#             self.queue.put(None)
#             self.stream.join()
#
#             transcripts = self.stream.transcriptions
#
#             self.stream = None
#             self.queue = None
#             return transcripts
#         return None
#
#     def execute(self, audio, language=None):
#         if self.server_bus:
#             start_time = time()
#             transcripts = self.stream_stop()
#             transcribe_time = time() - start_time
#             stt_name = repr(self.__class__.__name__)
#             print(f"{stt_name} | time={transcribe_time}")
#             self.server_bus.emit(Message("neon.metric", {"name": "stt execute",
#                                                          "transcripts": transcripts,
#                                                          "time": transcribe_time,
#                                                          "module": stt_name}))
#             return transcripts
#         else:
#             return self.stream_stop()
#
#     @abstractmethod
#     def create_streaming_thread(self):
#         pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
import time
from typing import Optional

from mycroft_bus_client import Message

from neon_utils.logger import LOG


def neon_must_respond(message: Message) -> bool:
    return True


def neon_in_request(message: Message) -> bool:
    return True


def request_from_mobile(message: Message) -> bool:
    if message.context.get("mobile"):
        return True
    return False


def preference_brands(message: Message) -> dict:
    return {'ignored_brands': {},
            'favorite_brands': {},
            'specially_requested': {}}


def preference_user(message: Message) -> dict:
    return {'first_name': '',
            'middle_name': '',
            'last_name': '',
            'preferred_name': '',
            'full_name': '',
            'dob': 'YYYY/MM/DD',
            'age': '',
            'email': '',
            'username': '',
            'password': '',
            'picture': '',
            'about': '',
            'phone': '',
            'email_verified': False,
            'phone_verified': False
            }


def preference_location(message: Message) -> dict:
    return {'lat': 47.4799078,
            'lng': -122.2034496,
            'city': 'Renton',
            'state': 'Washington',
            'country': 'USA',
            'tz': 'America/Los_Angeles',
            'utc': -8.0
            }


def preference_unit(message: Message) -> dict:
    return {'time': 12,
            'date': 'MDY',
            'measure': 'imperial'
            }


def preference_speech(message: Message) -> dict:
    return {'stt_language': 'en',
            'stt_region': 'US',
            'alt_languages': ['en'],
            'tts_language': "en-us",
            'tts_gender': 'female',
            'neon_voice': 'Joanna',
            'secondary_tts_language': '',
            'secondary_tts_gender': '',
            'secondary_neon_voice': '',
            'speed_multiplier': 1.0,
            'synonyms': {}
            }


def update_skill_settings(new_settings: dict, message: Message):
    from neon_utils import SKILL
    for pref, val in new_settings.items():
        SKILL.settings[pref] = val


def preference_skill(message: Message) -> dict:
    from neon_utils import SKILL
    return SKILL.settings


def build_user_dict(message: Message = None) -> dict:
    merged_dict = {**preference_speech(message), **preference_user(message),
                   **preference_brands(message), **preference_location(message),
                   **preference_unit(message)}
    for key, value in merged_dict.items():
        if value == "":
            merged_dict[key] = -1
    return merged_dict


def speak_dialog(key, data=None, expect_response=False, message=None, private=False, speaker=None, wait=False):
    """ Speak a random sentence from a dialog file.

    Arguments:
        :param key: dialog file key (e.g. "hello" to speak from the file "locale/en-us/hello.dialog")
        :param data: information used to populate key
        :param expect_response: set to True if Mycroft should listen for a response immediately after speaking.
        :param wait: set to True to block while the text is being spoken.
        :param speaker: optional dict of speaker info to use
        :param private: private flag (server use only)
        :param message: associated message from request
    """
    from neon_utils import SKILL, TYPE
    super(TYPE, SKILL).speak_dialog(key, data, expect_response, wait)


def create_signal(signal_name: str) -> bool:
    """Create a named signal. i.e. "CORE_signalName" or "nick_SKILL_signalName
    Args:
        signal_name (str): The signal's name.  Must only contain characters
            valid in filenames.
    """
    try:
        path = os.path.join('/tmp/mycroft/ipc', "signal", signal_name)
        _create_file(path)
        return os.path.isfile(path)
    except IOError:
        return False


def clear_signals(prefix: str):
    """
    Clears all signals that begin with the passed prefix. Used with skill prefix for a skill to clear any signals it
    may have set
    :param prefix: (str) prefix to match
    """
    os.makedirs("/tmp/mycroft/ipc/signal", exist_ok=True)
    for signal in os.listdir("/tmp/mycroft/ipc/signal"):
        if str(signal).startswith(prefix) or f"_{prefix}_" in str(signal):
            os.remove(os.path.join("/tmp/mycroft/ipc/signal", signal))


def check_for_signal(signal_name: str, sec_lifetime: Optional[int] = 0):
    """See if a named signal exists

    Args:
        signal_name (str): The signal's name.  Must only contain characters
            valid in filenames.
        sec_lifetime (int, optional): How many seconds the signal should
            remain valid.  If 0 or not specified, it is a single-use signal.
            If -1, it never expires.

    Returns:
        bool: True if the signal is defined, False otherwise
    """
    import time
    path = os.path.join('/tmp/mycroft/ipc', "signal", signal_name)
    if os.path.isfile(path):
        # noinspection PyTypeChecker
        if sec_lifetime == 0:
            # consume this single-use signal
            try:
                os.remove(path)
            except Exception as x:
                print(' >>> ERROR removing signal ' + signal_name + ', error == ' + str(x))
        elif sec_lifetime == -1:
            return True
        elif int(os.path.getctime(path) + sec_lifetime) < int(time.time()):
            # remove once expired
            os.remove(path)
            return False
        return True
    # No such signal exists
    return False


def _create_file(filename: str):
    """ Create the file filename and create any directories needed

        Args:
            filename: Path to the file to be created
    """
    try:
        os.makedirs(os.path.dirname(filename))
    except OSError:
        pass
    with open(filename, 'w') as f:
        f.write('')


def request_check_timeout(time_wait, intent_to_check):
    pass


def get_utterance_user(message) -> str:
    return "local"


def get_cached_data(filename: str, file_loc: Optional[str] = None) -> dict:
    """
    Retrieves cache data from a file created/updated with update_cached_data
    :param filename: (str) filename of cache object to update
    :param file_loc: (str) path to directory containing filename (defaults to cache dir)
    :return: (dict) cache data
    """
    import pickle
    import pathlib

    from neon_utils import SKILL

    if not file_loc:
        file_loc = SKILL.cache_loc
    cached_location = os.path.join(file_loc, filename)
    if pathlib.Path(cached_location).exists():
        with open(cached_location, 'rb') as file:
            return pickle.load(file)
    else:
        return {}


def update_cached_data(filename: str, new_element):
    """
    Updates cache file of skill responses to translated responses when non-english responses are requested.
    :param filename: (str) filename of cache object to update (relative to cacheDir)
    :param new_element: (any) object to cache at passed location
    """
    import pickle

    from neon_utils import SKILL

    with open(os.path.join(SKILL.cache_loc, filename), 'wb+') as file_to_update:
        pickle.dump(new_element, file_to_update, protocol=pickle.HIGHEST_PROTOCOL)


def build_message(kind, utt, message, signal_to_check=None, speaker=None):
    """
    Build a message for user input or neon response
    :param kind: "neon speak" or "execute"
    :param utt: string to emit
    :param message: incoming message object
    :param signal_to_check: signal to check in speech
    :param speaker: speaker data dictionary
    :return: Message object
    """
    from copy import deepcopy

    from neon_utils import SKILL
    # utt = utt.strip('"')  This is done before calling build_message now

    # Use utt as default signal to check
    if not signal_to_check:
        signal_to_check = utt
    LOG.debug(speaker)

    default_speech = SKILL.preference_speech(message)
    # Override user preference for all script responses
    if not speaker:
        speaker = {"name": "Neon",
                   "language": default_speech["tts_language"],
                   "gender": default_speech["tts_gender"],
                   "voice": default_speech["neon_voice"],
                   "override_user": True}
    else:
        speaker["override_user"] = True

    LOG.debug(f"data={message.data}")
    LOG.debug(f"context={message.context}")

    emit_response = False
    if kind == "skill_data":
        emit_response = True
        kind = "execute"

    try:
        if kind in ("execute", "skill"):
            message.context["cc_data"] = message.context.get("cc_data", {})
            # This is picked up in the intent handler
            # return message.reply("skills:execute.utterance", {
            return message.reply("recognizer_loop:utterance", {
                "utterances": [utt.lower()],
                "lang": message.data.get("lang", "en-US"),
                "session": None,
                "ident": None,
                "speaker": speaker
            }, {
                # "mobile": message.context.get("mobile", False),
                # "client": message.context.get("client", None),
                # "flac_filename": message.context.get("flac_filename", ''),
                # "nick_profiles": message.context.get("nick_profiles", {}),
                "neon_should_respond": True,
                "cc_data": {"signal_to_check": signal_to_check,
                            "request": utt,
                            "emit_response": emit_response,
                            # "Neon": True,
                            "execute_from_script": True,
                            "audio_file": message.context["cc_data"].get("audio_file", None),
                            "raw_utterance": utt
                            }
            })
        elif kind == "neon speak":
            context = deepcopy(message.context)
            LOG.info(f"CONTEXT IS {context}")
            context["cc_data"] = context.get("cc_data", {})
            context["cc_data"]["signal_to_check"] = signal_to_check
            context["cc_data"]["request"] = utt

            return message.reply("speak", {"lang": message.data.get("lang", "en-US"),
                                           "speaker": speaker
                                           }, context)
    except Exception as x:
        LOG.error(x)


def to_system_time(dt):
    """
    Converts a timezone aware datetime object to an object to be used in the messagebus scheduler
    :param dt: datetime object to convert
    :return: timezone aware datetime object that can be scheduled
    """
    from dateutil.tz import tzlocal, gettz

    tz = tzlocal()
    if dt.tzinfo:
        return dt.astimezone(tz)
    else:
        return dt.replace(tzinfo=gettz("UTC")).astimezone(tz)


def speak(utterance, expect_response=False, wait=False, meta=None, message=None, private=False, speaker=None):
    """
    Speak a sentence.
    Arguments:
        utterance (str):        sentence mycroft should speak
        expect_response (bool): set to True if Mycroft should listen for a response immediately after
                                speaking the utterance.
        message (Message):      message associated with the input that this speak is associated with
        private (bool):         flag to indicate this message contains data that is private to the requesting user
        speaker (dict):         dict containing language or voice data to override user preference values
        wait (bool):            set to True to block while the text is being spoken.
        meta:                   Information of what built the sentence.
    """
    from neon_utils import SKILL

    # registers the skill as being active
    meta = meta or {}
    meta['skill'] = SKILL.name
    SKILL.enclosure.register(SKILL.name)
    if utterance:
        LOG.debug(f">>>>> Skill speak! {utterance}")

        # Find the associated message
        if message:
            LOG.info('message passed to speak = ' + str(message.data))
            if not speaker:
                speaker = message.data.get("speaker", None)
        else:
            LOG.debug('message is None.')
            message = dig_for_message()

        if message:
            # filename = message.context.get("flac_filename", "")
            # cc_data = message.context.get("cc_data", {})
            # profiles = message.context.get("nick_profiles", {})
            if not speaker:
                speaker = message.data.get("speaker", speaker)
            # if message.data['flac_filename']:
            #     filename = message.data['flac_filename']
            # else:
            #     filename = ''
        else:
            message = dig_for_message()
            filename = ''
            # cc_data = {}
            # profiles = {}
            if message:
                # filename = message.context.get("flac_filename", "")
                # cc_data = message.context.get("cc_data", {})
                # profiles = message.context.get("nick_profiles", {})
                if not speaker:
                    speaker = message.data.get("speaker", {})

        # registers the skill as being active
        # print(f'{cc_data} is cc_data')
        # self.enclosure.register(self.name)
        nick = ""
        # LOG.debug(nick)
        data = {"utterance": utterance,
                "expect_response": expect_response,
                "meta": meta,
                "speaker": speaker}

        # devices might not want to do these logs either... weird characters cause a logging error
        if not SKILL.server:
            LOG.info(f'{speaker} Speak: {utterance}')
            # LOG.info('Speak data = ' + str(data))
        # LOG.info(filename)
        if not message:
            message = dig_for_message()

        if message and message.context.get("cc_data", {}).get("emit_response"):
            LOG.debug(f"DM: {data}")
            msg_to_emit = message.reply("skills:execute.response", data)

        elif message and message.msg_type != "mycroft.ready":
            message.context.get("timing", {})["speech_start"] = time.time()
            LOG.info("message True, " + str(data))
            # LOG.info(message)
            # TODO: This is where we have the most complete timing profile for an utterance
            # LOG.debug(f"TIME: to_speak, {time.time()}, {message.context['flac_filename']}, {data['utterance']}, "
            #           f"{message.context}")
            # self.bus.emit(message.reply("speak", data))
            msg_to_emit = message.reply("speak", data)
            LOG.debug(f">>>> Skill speak! {data}, {message.context}")
        else:
            LOG.warning("message False, " + str(data))
            # self.bus.emit(Message("speak", data))
            msg_to_emit = Message("speak", data)
        LOG.debug(msg_to_emit.msg_type)
        SKILL.bus.emit(msg_to_emit)
    else:
        LOG.warning("Null utterance passed to speak")
        LOG.warning(f"{SKILL.name} | message={message}")

    if wait:
        wait_while_speaking()


def dig_for_message():
    """Dig Through the stack for message."""
    import inspect

    stack = inspect.stack()
    # Limit search to 10 frames back
    stack = stack if len(stack) < 10 else stack[:10]
    local_vars = [frame[0].f_locals for frame in stack]
    for var in local_vars:
        if 'message' in var and isinstance(var['message'], Message):
            return var['message']


# TODO: Are the below methods from elsewhere in core (not skills)? DM
def wait_while_speaking():
    """Pause as long as Text to Speech is still happening

    Pause while Text to Speech is still happening.  This always pauses
    briefly to ensure that any preceding request to speak has time to
    begin.
    """
    LOG.debug("Wait while speaking!")
    time.sleep(0.3)  # Wait briefly in for any queued speech to begin
    while is_speaking():
        time.sleep(0.1)


def is_speaking(sec_lifetime=-1):
    """Determine if Text to Speech is occurring

    Args:
        sec_lifetime (int, optional): How many seconds the signal should
            remain valid.  If 0 or not specified, it is a single-use signal.
            If -1, it never expires.

    Returns:
        bool: True while still speaking
    """
    return check_for_signal("isSpeaking", sec_lifetime)


def check_yes_no_response(message):
    """
    Used in converse methods to check if a response confirms or declines an action. Differs from ask_yesno in that
    this does not assume input will be spoken with wake words enabled
    :param message: incoming message object to evaluate
    :return: False if declined, numbers if confirmed numerically, True if confirmed with no numbers
    """
    from neon_utils import SKILL

    utterance = message.data.get("utterances")[0]
    if SKILL.voc_match(utterance, "no"):
        LOG.info("User Declined")
        return False
    elif SKILL.voc_match(utterance, "yes"):
        LOG.info("User Accepted")
        numbers = [str(s) for s in utterance.split() if s.isdigit()]
        if numbers and len(numbers) > 0:
            confirmation = "".join(numbers)
            LOG.info(f"Got confirmation: {confirmation}")
            return confirmation
        return True
    else:
        LOG.debug("User response not valid")
        return -1



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import sys
import re
import importlib.util
import pkg_resources
import sysconfig

from os.path import exists, join


def get_packaged_core_version() -> str:
    """
    Get the version of the packaged core in use. Supports Neon, Mycroft, and OVOS default packages.
    Returns:
        Version of the installed core package
    """
    if importlib.util.find_spec("neon-core"):
        return pkg_resources.get_distribution("neon-core")
    elif importlib.util.find_spec("mycroft-core"):
        return pkg_resources.get_distribution("mycroft-core")
    elif importlib.util.find_spec("mycroft-lib"):
        return pkg_resources.get_distribution("mycroft-lib")
    raise ImportError("No Core Package Found")


def get_version_from_file() -> str:
    """
    Get the core version from a .release file, Provides legacy support for Neon
    Returns:
        Version of the cloned core repository
    """
    import glob
    import os
    from neon_utils.configuration_utils import get_neon_local_config
    release_files = glob.glob(
        f'{get_neon_local_config().get("dirVars", {}).get("ngiDir") or os.path.expanduser("~/.neon")}'
        f'/*.release')
    if len(release_files):
        return os.path.splitext(os.path.basename(release_files[0]))[0]
    raise FileNotFoundError("No Version File Found")


def get_neon_core_version() -> str:
    """
    Gets the current version of the installed Neon Core.
    Returns:
        Version of the available/active Neon Core or 0.0 if no release info is found
    """
    try:
        return get_packaged_core_version()
    except ImportError:
        pass

    try:
        return get_version_from_file()
    except FileNotFoundError:
        pass

    return "0.0"


def get_core_root():
    """
    Determines the root of the available/active Neon Core. Should be the immediate parent directory of 'mycroft' dir
    Returns:
        Path to the core directory containing 'mycroft'
    """
    site = sysconfig.get_paths()['platlib']
    if exists(join(site, 'mycroft')):
        return site
    for p in [path for path in sys.path if path != ""]:
        if exists(join(p, "mycroft")):
            return p
        if re.match(".*/lib/python.*/site-packages", p):
            clean_path = "/".join(p.split("/")[0:-4])
            if exists(join(clean_path, "mycroft")):
                return clean_path
            # TODO: Other packages (Neon Core)? DM
            elif exists(join(p, "mycroft")):
                return p
    raise FileNotFoundError("Could not determine core directory")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
import boto3
from neon_utils.logger import LOG
from neon_utils.configuration_utils import get_neon_lang_config, NGIConfig, get_neon_tts_config


def get_language_dir(base_path, lang="en-us"):
    """ checks for all language variations and returns best path """
    lang_path = os.path.join(base_path, lang)
    # base_path/en-us
    if os.path.isdir(lang_path):
        return lang_path
    if "-" in lang:
        main = lang.split("-")[0]
        # base_path/en
        general_lang_path = os.path.join(base_path, main)
        if os.path.isdir(general_lang_path):
            return general_lang_path
    else:
        main = lang
    # base_path/en-uk, base_path/en-au...
    if os.path.isdir(base_path):
        candidates = [os.path.join(base_path, f)
                      for f in os.listdir(base_path) if f.startswith(main)]
        paths = [p for p in candidates if os.path.isdir(p)]
        # TODO how to choose best local dialect?
        if len(paths):
            return paths[0]
    return os.path.join(base_path, lang)


class LanguageDetector:
    def __init__(self):
        self.config = get_neon_lang_config()
        self.default_language = self.config["user"].split("-")[0]
        # hint_language: str  E.g., 'ITALIAN' or 'it' boosts Italian
        self.hint_language = self.config["user"].split("-")[0]
        self.boost = self.config.get("boost")

    def detect(self, text):
        # assume default language
        return self.default_language

    def detect_probs(self, text):
        return {self.detect(text): 1}


class LanguageTranslator:
    def __init__(self):
        self.config = get_neon_lang_config()
        self.boost = self.config["boost"]
        self.default_language = self.config["user"].split("-")[0]
        self.internal_language = self.config["internal"]

    def translate(self, text, target=None, source=None):
        return text


class GoogleDetector(LanguageDetector):
    def __init__(self):
        super().__init__()
        from ovos_utils.lang.detect import detect_lang
        import logging
        logging.getLogger("hyper.http20.connection").setLevel("ERROR")
        self._detect = detect_lang

    def detect(self, text):
        return self._detect(text) or self.default_language

    def detect_probs(self, text):
        lang = self._detect(text, True)
        return {lang["lang_code"]: lang["conf"]}


class FastLangDetector(LanguageDetector):
    def __init__(self):
        super().__init__()
        try:
            from fastlang import fastlang
        except ImportError:
            LOG.error("Run pip install fastlang")
            raise
        self._detect = fastlang

    def detect(self, text):
        return self._detect(text)["lang"]

    def detect_probs(self, text):
        return self._detect(text)["probabilities"]


class LangDetectDetector(LanguageDetector):
    def __init__(self):
        super().__init__()
        try:
            from langdetect import detect, detect_langs
        except ImportError:
            LOG.error("Run pip install langdetect")
            raise
        self._detect = detect
        self._detect_prob = detect_langs

    def detect(self, text):
        return self._detect(text)

    def detect_probs(self, text):
        langs = {}
        for lang in self._detect_prob(text):
            langs[lang.lang] = lang.prob
        return langs


class GoogleTranslator(LanguageTranslator):
    def __init__(self):
        super().__init__()
        from ovos_utils.lang.translate import translate_text
        self._translate = translate_text

    def translate(self, text, target=None, source=None):
        if self.boost and not source:
            source = self.default_language
        target = target or self.internal_language
        return self._translate(text, target, source)


class AmazonTranslator(LanguageTranslator):
    def __init__(self):
        super().__init__()
        # TODO: Replace private key function DM
        # self.keys = get_private_keys()["amazon"]
        self.keys = get_neon_tts_config()["amazon"]
        self.client = boto3.Session(aws_access_key_id=self.keys["aws_access_key_id"],
                                    aws_secret_access_key=self.keys["aws_secret_access_key"],
                                    region_name=self.keys["region"]).client('translate')

    def translate(self, text, target=None, source="auto"):
        target = target or self.internal_language

        response = self.client.translate_text(
            Text=text,
            SourceLanguageCode=source,
            TargetLanguageCode=target.split("-")[0]
        )
        return response["TranslatedText"]


class AmazonDetector(LanguageDetector):
    def __init__(self):
        super().__init__()
        # TODO: Replace private key funciton DM
        # self.keys = get_private_keys()["amazon"]
        self.keys = get_neon_tts_config()["amazon"]

        self.client = boto3.Session(aws_access_key_id=self.keys["aws_access_key_id"],
                                    aws_secret_access_key=self.keys["aws_secret_access_key"],
                                    region_name=self.keys["region"]).client('comprehend')

    def detect(self, text):
        response = self.client.detect_dominant_language(
            Text=text
        )
        return response['Languages'][0]['LanguageCode']

    def detect_probs(self, text):
        response = self.client.detect_dominant_language(
            Text=text
        )
        langs = {}
        for lang in response["Languages"]:
            langs[lang["LanguageCode"]] = lang["Score"]
        return langs


class MyMemoryTranslator(LanguageTranslator):
    def __init__(self):
        super().__init__()

    def translate(self, text, target=None, source="auto"):
        from translate import Translator
        target = target or self.internal_language
        # TODO: This doesn't appear to work DM
        translated = Translator(to_lang=target).translate(text)
        LOG.info(translated)
        return translated


class TranslatorFactory:
    CLASSES = {
        "google": GoogleTranslator,
        "amazon": AmazonTranslator,
        # "mymemory": MyMemoryTranslator
    }

    @staticmethod
    def create(module=None):
        module = module or "amazon"
        config = get_neon_lang_config()
        module = module or config.get("translation_module", "google")
        # TODO: Check file locations DM
        if module == "amazon" \
                and get_neon_tts_config()["amazon"].get("aws_access_key_id", "") == "":
            from neon_utils.authentication_utils import find_neon_aws_keys, populate_amazon_keys_config
            try:
                config_keys = find_neon_aws_keys()
                populate_amazon_keys_config(config_keys)
            except FileNotFoundError:
                LOG.debug("Amazon credentials not available")
                module = "google"
            except Exception as e:
                LOG.error(e)
                module = "google"
        try:
            clazz = TranslatorFactory.CLASSES.get(module)
            return clazz()
        except Exception as e:
            # The translate backend failed to start. Report it and fall back to
            # default.
            LOG.exception('The selected translator backend could not be loaded, '
                          'falling back to default...')
            LOG.error(e)
            if module != 'google':
                return GoogleTranslator()
            else:
                raise


class DetectorFactory:
    CLASSES = {
        "amazon": AmazonDetector,
        "google": GoogleDetector,
        "fastlang": FastLangDetector,
        "detect": LangDetectDetector
    }

    @staticmethod
    def create(module=None):
        module = module or "fastlang"
        config = get_neon_lang_config()
        module = module or config.get("detection_module", "fastlang")
        try:
            clazz = DetectorFactory.CLASSES.get(module)
            return clazz()
        except Exception as e:
            # The translate backend failed to start. Report it and fall back to
            # default.
            LOG.exception('The selected language detector backend could not be loaded, '
                          'falling back to default...')
            LOG.error(e)
            if module != 'fastlang':
                return FastLangDetector()
            else:
                raise



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from time import time


class Stopwatch:
    """
    Provides a stopwatch object compatible with mycroft.metrics Stopwatch.
    """
    def __init__(self, metric_name=None, allow_reporting=False):
        """
        Create a stopwatch object with an optional metric_name
        Args:
            metric_name: name of the metric this stopwatch is measuring
            allow_reporting: boolean flag to allow this stopwatch to report measured metrics
        """
        self._metric = metric_name
        self._report = allow_reporting
        self._context = dict()
        self.start_time = None
        self.time = None

    def __enter__(self):
        self.start()

    def __exit__(self, typ, val, traceback):
        self.stop()
        # self.report()

    # def add_context(self, context: dict):
    #     """
    #     Add context to the measured metric.
    #     Args:
    #         context: dict of arbitrary data to add to this metric reporting
    #     """
    #     if self._metric:
    #         self._context = context

    def start(self):
        self.start_time = time()

    def stop(self):
        self.time = time() - self.start_time
        return self.time

    # def report(self):
    #     if self._metric and self._report:
    #         report_metric(self._metric, self._context)
    #         self._context = dict()


# def report_metric(name: str, data: dict, config: dict = None):
#     """
#     Report a metric via the messagebus.
#     Args:
#         name: Name of the metric to report
#         data: Arbitrary metric data to report
#         config: Optional messagebus config
#     """
#     combined = deepcopy(data)
#     combined['name'] = name
#     mycroft_bus_client.send("neon.metric", combined, config)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import json
import time
from collections import OrderedDict


class LRUCache:
    # TODO user specific cache with a compound key
    def __init__(self, capacity: int = 128):
        self.cache = OrderedDict()
        self._capacity = capacity
        self._hits = 0
        self._missed = 0
        self._init_time = time.time()

    def __len__(self):
        return len(self.cache)

    @property
    def hits(self):
        return self._hits

    @property
    def missed(self):
        return self._missed

    def get(self, key):
        """
        Return the value of the key from cache.
        Increment self._missed in key not in cache, increment self._hits in key in cache
        Args:
            key: a key to look up in cache
        Returns: value associated with the key or None

        """
        if key not in self.cache:
            self._missed += 1
            return None
        else:
            self._hits += 1
            self.cache.move_to_end(key)
            return self.cache[key]

    def put(self, key, value) -> None:
        """
        Put a key-value pair into cache
        Args:
            key: a key to put into cache
            value: a value to put into cache
        Returns:
        """
        self.cache[key] = value
        self.cache.move_to_end(key)
        if len(self.cache) > self._capacity:
            self.cache.popitem(last=False)

    def clear(self):
        """
        Clear the cache dict
        """
        self.cache = OrderedDict()

    def clear_full(self):
        """
        Clear the cache dict, reset hits, missed and initialization time
        Returns:

        """
        self.clear()
        self._hits = self._missed = 0
        self._init_time = time.time()

    def jsonify(self):
        """
        Dump cache dict into json
        Returns: json-string
        """
        return json.dumps(self.cache)

    def jsonify_metrics(self):
        """
        Dump cache metrics into json
        Returns:

        """
        data = dict()
        data["cache"] = self.cache
        data["hits"] = self.hits
        data["missed"] = self.missed
        data["size"] = self._capacity
        return json.dumps(data)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import socket
import netifaces


def get_ip_address() -> str:
    """
    Returns the IPv4 address of the default interface (This is a public IP for server implementations)
    :return: IP Address
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    return s.getsockname()[0]


def get_adapter_info(interface: str = "default") -> dict:
    """
    Returns MAC and IP info for the specified gateway
    :param interface: Name of network interface to check
    :return: Dict of mac, IPv4 and IPv6 addresses
    """
    gateways = netifaces.gateways()
    if interface not in gateways:
        raise IndexError("Requested Interface Not Found")
    if netifaces.AF_INET not in gateways[interface]:
        raise IndexError("Requested Interface not connected!")
    device = gateways[interface][netifaces.AF_INET][1]
    mac = netifaces.ifaddresses(device)[netifaces.AF_LINK][0]['addr']
    ip4 = netifaces.ifaddresses(device)[netifaces.AF_INET][0]['addr']
    ip6 = netifaces.ifaddresses(device)[netifaces.AF_INET6][0]['addr']
    return {"ipv4": ip4, "ipv6": ip6, "mac": mac}



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
#
# This software is an enhanced derivation of the Mycroft Project which is licensed under the
# Apache software Foundation software license 2.0 https://www.apache.org/licenses/LICENSE-2.0
# Changes Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum
from abc import ABC, abstractmethod
from neon_utils.skills.neon_skill import NeonSkill


class CQSMatchLevel(IntEnum):
    EXACT = 1  # Skill could find a specific answer for the question
    CATEGORY = 2  # Skill could find an answer from a category in the query
    GENERAL = 3  # The query could be processed as a general query


# Copy of CQSMatchLevel to use if the skill returns visual media
CQSVisualMatchLevel = IntEnum('CQSVisualMatchLevel',
                              [e.name for e in CQSMatchLevel])


def is_CQSVisualMatchLevel(match_level):
    return isinstance(match_level, type(CQSVisualMatchLevel.EXACT))


VISUAL_DEVICES = ['mycroft_mark_2']


def handles_visuals(platform):
    return platform in VISUAL_DEVICES


class CommonQuerySkill(NeonSkill, ABC):
    """Question answering skills should be based on this class.

    The skill author needs to implement `CQS_match_query_phrase` returning an
    answer and can optionally implement `CQS_action` to perform additional
    actions if the skill's answer is selected.

    This class works in conjunction with skill-query which collects
    answers from several skills presenting the best one available.
    """
    def __init__(self, name=None, bus=None):
        super().__init__(name, bus)

    def bind(self, bus):
        """Overrides the default bind method of MycroftSkill.

        This registers messagebus handlers for the skill during startup
        but is nothing the skill author needs to consider.
        """
        if bus:
            super().bind(bus)
            self.add_event('question:query', self.__handle_question_query)
            self.add_event('question:action', self.__handle_query_action)

    def __handle_question_query(self, message):
        search_phrase = message.data["phrase"]

        # First, notify the requestor that we are attempting to handle
        # (this extends a timeout while this skill looks for a match)
        self.bus.emit(message.response({"phrase": search_phrase,
                                        "skill_id": self.skill_id,
                                        "searching": True}))

        # Now invoke the CQS handler to let the skill perform its search
        result = self.CQS_match_query_phrase(search_phrase, message)

        if result:
            match = result[0]
            level = result[1]
            answer = result[2]
            callback = result[3] if len(result) > 3 else None
            confidence = self.__calc_confidence(match, search_phrase, level)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "answer": answer,
                                            "callback_data": callback,
                                            "conf": confidence}))
        else:
            # Signal we are done (can't handle it)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "searching": False}))

    def __calc_confidence(self, match, phrase, level):
        # Assume the more of the words that get consumed, the better the match
        consumed_pct = len(match.split()) / len(phrase.split())
        if consumed_pct > 1.0:
            consumed_pct = 1.0

        # Add bonus if match has visuals and the device supports them.
        # platform = self.config_core.get('enclosure', {}).get('platform')
        # if is_CQSVisualMatchLevel(level) and handles_visuals(platform):
        #     bonus = 0.1
        # else:
        #     bonus = 0
        bonus = 0

        if int(level) == int(CQSMatchLevel.EXACT):
            return 0.9 + (consumed_pct / 10) + bonus
        elif int(level) == int(CQSMatchLevel.CATEGORY):
            return 0.6 + (consumed_pct / 10) + bonus
        elif int(level) == int(CQSMatchLevel.GENERAL):
            return 0.5 + (consumed_pct / 10) + bonus
        else:
            return 0.0  # should never happen

    def __handle_query_action(self, message):
        """Message handler for question:action.

        Extracts phrase and data from message forward this to the skills
        CQS_action method.
        """
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        phrase = message.data["phrase"]
        data = message.data.get("callback_data")
        # Invoke derived class to provide playback data
        self.CQS_action(phrase, data)

    @abstractmethod
    def CQS_match_query_phrase(self, phrase, message):
        """Analyze phrase to see if it is a play-able phrase with this skill.

        Needs to be implemented by the skill.

        Arguments:
            phrase (str): User phrase, "What is an aardwark"

        Returns:
            (match, CQSMatchLevel[, callback_data]) or None: Tuple containing
                 a string with the appropriate matching phrase, the PlayMatch
                 type, and optionally data to return in the callback if the
                 match is selected.
        """
        # Derived classes must implement this, e.g.
        return None

    def CQS_action(self, phrase, data):
        """Take additional action IF the skill is selected.

        The speech is handled by the common query but if the chosen skill
        wants to display media, set a context or prepare for sending
        information info over e-mail this can be implemented here.

        Args:
            phrase (str): User phrase uttered after "Play", e.g. "some music"
            data (dict): Callback data specified in match_query_phrase()
        """
        # Derived classes may implement this if they use additional media
        # or wish to set context after being called.
        pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
from datetime import datetime

import pendulum
from dateutil.tz import tzlocal, gettz

from geopy.geocoders import Nominatim
from timezonefinder import TimezoneFinder

from neon_utils import LOG


def get_coordinates(gps_loc: dict) -> (float, float):
    """
    Gets the latitude and longitude for the passed location
    :param gps_loc: dict of "city", "state", "country"
    :return: lat, lng float values
    """
    coordinates = Nominatim(user_agent="neon-ai")
    try:
        location = coordinates.geocode(gps_loc)
        LOG.debug(f"{location}")
        return location.latitude, location.longitude
    except Exception as x:
        LOG.error(x)
        return -1, -1


def get_location(lat, lng) -> (str, str, str, str):
    """
    Gets location name values for the passed coordinates.
    Note that some coordinates do not have a city, but may have a county.
    :param lat: latitude
    :param lng: longitude
    :return: city, county, state, country
    """
    address = Nominatim(user_agent="neon-ai")
    location = address.reverse([lat, lng], language="en-US")
    LOG.debug(f"{location}")
    LOG.debug(f"{location.raw}")
    LOG.debug(f"{location.raw.get('address')}")
    city = location.raw.get('address').get('city') or location.raw.get('address').get('town')
    county = location.raw.get('address').get('county')
    state = location.raw.get('address').get('state')
    country = location.raw.get('address').get('country')
    return city, county, state, country


def get_timezone(lat, lng) -> (str, float):
    """
    Gets timezone information for the passed coordinates.
    Note that some coordinates do not have a city, but may have a county.
    :param lat: latitude
    :param lng: longitude
    :return: timezone name, offset from GMT
    """
    timezone = TimezoneFinder().timezone_at(lng=float(lng), lat=float(lat))
    offset = pendulum.from_timestamp(0, timezone).offset_hours
    return timezone, offset


def to_system_time(dt: datetime) -> datetime:
    """
    Converts a timezone aware or timezone naiive datetime object to a datetime object in the system tz
    :param dt: datetime object to convert
    :return: timezone aware datetime object that can be scheduled
    """
    tz = tzlocal()
    if dt.tzinfo:
        return dt.astimezone(tz)
    else:
        return dt.replace(tzinfo=tz).astimezone(tz)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import glob
import os
import base64
import wave
from tempfile import mkstemp

from typing import Optional
from pydub import AudioSegment
from ovos_utils.signal import ensure_directory_exists

from neon_utils import LOG


def encode_file_to_base64_string(path: str) -> str:
    """
    Encodes a file to a base64 string (useful for passing file data over a messagebus)
    :param path: Path to file to be encoded
    :return: encoded string
    """
    if not isinstance(path, str):
        raise TypeError
    path = os.path.expanduser(path)
    if not os.path.isfile(path):
        LOG.error(f"File Not Found: {path}")
        raise FileNotFoundError
    with open(path, "rb") as file_in:
        encoded = base64.b64encode(file_in.read()).decode("utf-8")
    return encoded


def decode_base64_string_to_file(encoded_string: str, output_path: str) -> str:
    """
    Writes out a base64 string to a file object at the specified path
    :param encoded_string: Base64 encoded string
    :param output_path: Path to file to write (throws exception if file exists)
    :return: Path to output file
    """
    if not isinstance(output_path, str):
        raise TypeError
    output_path = os.path.expanduser(output_path)
    if os.path.isfile(output_path):
        LOG.error(f"File already exists: {output_path}")
        raise FileExistsError
    ensure_directory_exists(os.path.dirname(output_path))
    with open(output_path, "wb+") as file_out:
        byte_data = base64.b64decode(encoded_string.encode("utf-8"))
        file_out.write(byte_data)
    return output_path


def get_most_recent_file_in_dir(path: str, ext: Optional[str] = None) -> Optional[str]:
    """
    Gets the most recently created file in the specified path
    :param path: File path or glob pattern to check
    :param ext: File extension
    :return: Path to newest file in specified path with specified extension (None if no files in path)
    """
    if os.path.isdir(path):
        path = f"{path}/*"
    list_of_files = glob.glob(path)  # * means all if need specific format then *.csv
    if ext:
        if not ext.startswith("."):
            ext = f".{ext}"
        list_of_files = [file for file in list_of_files if os.path.splitext(file)[1] == ext]
    if list_of_files:
        latest_file = max(list_of_files, key=os.path.getctime)
        return latest_file
    else:
        return None


def get_file_as_wav(audio_file: str, desired_sample_rate: int, desired_sample_width: int = 2) -> wave.Wave_read:
    """
    Gets a wav file for the passed audio file.
    Args:
        audio_file: Path to audio file in arbitrary format
        desired_sample_rate: sample rate at which returned wav data should be encoded
        desired_sample_width: sample width at which returned wav data should be encoded
    Returns:
        Wave_read object encoded at the desired_sample_rate
    """

    audio_file = os.path.expanduser(audio_file)
    if not os.path.isfile(audio_file):
        raise FileNotFoundError
    try:
        file = wave.open(audio_file, 'rb')
        sample_rate = file.getframerate()
        sample_width = file.getsampwidth()
        if sample_rate == desired_sample_rate and sample_width == desired_sample_width:
            return file
    except wave.Error:
        pass
    except Exception as e:
        raise e
    audio = AudioSegment.from_file(audio_file)
    audio = audio.set_frame_rate(desired_sample_rate).set_sample_width(desired_sample_width)
    _, tempfile = mkstemp()
    audio.export(tempfile, format='wav').close()
    return wave.open(tempfile, 'rb')


def get_audio_file_stream(wav_file: str, sample_rate: int = 16000):
    """
    Creates a FileStream object for the specified wav_file with the specified output sample_rate.
    Args:
        wav_file: Path to file to read
        sample_rate: Desired output sample rate (None for wav_file sample rate)

    Returns:
        FileStream object for the passed audio file
    """
    class FileStream:
        MIN_S_TO_DEBUG = 5.0

        # How long between printing debug info to screen
        UPDATE_INTERVAL_S = 1.0

        def __init__(self, file_name):
            self.file = get_file_as_wav(file_name, sample_rate)
            self.sample_rate = self.file.getframerate()
            # if sample_rate and self.sample_rate != sample_rate:
            #     sound = AudioSegment.from_file(file_name, format='wav', frame_rate=self.sample_rate)
            #     sound = sound.set_frame_rate(sample_rate)
            #     _, tempfile = mkstemp()
            #     sound.export(tempfile, format='wav')
            #     self.file = wave.open(tempfile, 'rb')
            #     self.sample_rate = self.file.getframerate()
            self.size = self.file.getnframes()
            self.sample_width = self.file.getsampwidth()
            self.last_update_time = 0.0

            self.total_s = self.size / self.sample_rate / self.sample_width

        def calc_progress(self):
            return float(self.file.tell()) / self.size

        def read(self, chunk_size):

            progress = self.calc_progress()
            if progress == 1.0:
                raise EOFError

            return self.file.readframes(chunk_size)

        def close(self):
            self.file.close()

    if not os.path.isfile(wav_file):
        raise FileNotFoundError
    try:
        return FileStream(wav_file)
    except Exception as e:
        raise e



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import pathlib
import pickle
import json
import shutil
import time
import os
import re

from copy import deepcopy
from functools import wraps
from itertools import chain
from mycroft_bus_client.message import Message, dig_for_message
from neon_utils.file_utils import get_most_recent_file_in_dir
from ruamel.yaml.comments import CommentedMap
from typing import Optional
from dateutil.tz import gettz
from neon_utils import create_signal, check_for_signal, wait_while_speaking
from neon_utils.configuration_utils import NGIConfig
from neon_utils.location_utils import to_system_time
from neon_utils.language_utils import get_neon_lang_config, DetectorFactory, TranslatorFactory
from neon_utils.logger import LOG
from neon_utils.message_utils import request_from_mobile, get_message_user
from neon_utils.cache_utils import LRUCache

from mycroft.skills.mycroft_skill.mycroft_skill import MycroftSkill
from mycroft.skills.skill_data import read_vocab_file

LOG.name = "neon_skill"


# TODO if accepted, make changes to QASkill and WikipediaSkill
# Available_modes are 1) quick; 2) thoughtful.
SPEED_MODE_EXTENSION_TIME = {
    "quick": 1,
    "thoughtful": 10
}
DEFAULT_SPEED_MODE = "thoughtful"
CACHE_TIME_OFFSET = 24*60*60  # seconds in 24 hours


class NeonSkill(MycroftSkill):
    def __init__(self, name=None, bus=None, use_settings=True):
        self.user_config = NGIConfig("ngi_user_info")
        self.local_config = NGIConfig("ngi_local_conf")

        self.ngi_settings: Optional[NGIConfig] = None

        super(NeonSkill, self).__init__(name, bus, use_settings)

        self.cache_loc = os.path.expanduser(self.local_config.get('dirVars', {}).get('cacheDir') or
                                            "~/.local/share/neon/cache")
        self.lru_cache = LRUCache()

        # TODO: Depreciate these references, signal use is discouraged DM
        self.create_signal = create_signal
        self.check_for_signal = check_for_signal

        self.sys_tz = gettz()
        self.gui_enabled = self.configuration_available.get("prefFlags", {}).get("guiEvents", False)

        if use_settings:
            self.settings = {}
            self._initial_settings = None
            self.init_settings()
        else:
            LOG.error(f"{name} Skill requested no settings!")
            self.settings = None

        self.scheduled_repeats = []

        # Server-specific imports and timeout setting
        # A server is a device that hosts the core and skills to serve clients,
        # but that a user will not interact with directly.
        # A server will likely serve multiple users and devices concurrently.
        if self.configuration_available.get("devVars", {}).get("devType", "generic") == "server":
            self.server = True
            self.default_intent_timeout = 90
        else:
            self.server = False
            self.default_intent_timeout = 60

        self.neon_core = True
        self.actions_to_confirm = dict()

        self.skill_mode = self.user_config.content.get('response_mode').get('speed_mode') or DEFAULT_SPEED_MODE
        self.extension_time = SPEED_MODE_EXTENSION_TIME.get(self.skill_mode)

        try:
            # Lang support
            self.language_config = get_neon_lang_config()
            self.lang_detector = DetectorFactory.create()  # Default fastlang
            self.translator = TranslatorFactory.create()  # Default Amazon
        except Exception as e:
            LOG.error(e)
            self.language_config, self.language_detector, self.translator = None, None, None

    def initialize(self):
        # schedule an event to load the cache on disk every CACHE_TIME_OFFSET seconds
        self.schedule_event(self._write_cache_on_disk, CACHE_TIME_OFFSET, name="neon.load_cache_on_disk")

    @property
    def user_info_available(self):
        LOG.warning("This reference is deprecated, use self.preference_x methods for user preferences")
        return self.user_config.content

    @property
    def configuration_available(self):
        LOG.warning("This reference is deprecated, use self.local_config directly")
        return self.local_config.content

    def init_settings(self):
        """
        Initializes yml-based skill config settings, updating from default dict as necessary for added parameters
        """
        if os.path.isfile(os.path.join(self.root_dir, "settingsmeta.yml")):
            skill_meta = NGIConfig("settingsmeta", self.root_dir).content
        elif os.path.isfile(os.path.join(self.root_dir, "settingsmeta.json")):
            with open(os.path.join(self.root_dir, "settingsmeta.json")) as f:
                skill_meta = json.load(f)
        else:
            skill_meta = None

        # Load defaults from settingsmeta
        default = {"__mycroft_skill_firstrun": True}
        if skill_meta:
            # LOG.info(skill_meta)
            LOG.info(skill_meta["skillMetadata"]["sections"])
            for section in skill_meta["skillMetadata"]["sections"]:
                for pref in section.get("fields", []):
                    if not pref.get("name"):
                        LOG.debug(f"non-data skill meta: {pref}")
                    else:
                        if pref.get("value") == "true":
                            value = True
                        elif pref.get("value") == "false":
                            value = False
                        elif isinstance(pref.get("value"), CommentedMap):
                            value = dict(pref.get("value"))
                        else:
                            value = pref.get("value")
                        default[pref["name"]] = value

        # Load or init configuration
        if os.path.isfile(os.path.join(self.root_dir, f"{self.name}.yml")):
            LOG.warning(f"Config found in skill directory for {self.name}! Relocating to: {self.file_system.path}")
            shutil.move(os.path.join(self.root_dir, f"{self.name}.yml"), self.file_system.path)
        self.ngi_settings = NGIConfig(self.name, self.file_system.path)

        # Load any new or updated keys
        try:
            LOG.debug(self.ngi_settings.content)
            LOG.debug(default)
            if self.ngi_settings.content and len(self.ngi_settings.content.keys()) > 0 and len(default.keys()) > 0:
                self.ngi_settings.make_equal_by_keys(default, recursive=False)
            elif len(default.keys()) > 0:
                LOG.info("No settings to load, use default")
                self.ngi_settings.populate(default)
        except Exception as e:
            LOG.error(e)
            self.ngi_settings.populate(default)

        # Make sure settings is initialized as a dictionary
        if self.ngi_settings.content:
            self.settings = self.ngi_settings.content  # Uses the default self.settings object for skills compat
        LOG.debug(f"loaded settings={self.settings}")

    @property
    def location_timezone(self):
        """Get the timezone code, such as 'America/Los_Angeles'"""
        LOG.warning("This method does not support user-specific location and will use device default")
        return self.preference_location()["tz"]

    def preference_brands(self, message=None) -> dict:
        """
        Returns a brands dictionary for the user
        Equivalent to self.user_config["speech"] for non-server use
        """
        try:
            nick = get_message_user(message) if message else None
            if self.server:
                if not message or not nick:
                    LOG.warning("No message given!")
                    return self.user_config['brands']

                if message.context.get("nick_profiles"):
                    return message.context["nick_profiles"][nick]["brands"]
                else:
                    LOG.error(f"Unable to get user settings! message={message.data}")
            else:
                return self.user_config['brands']
        except Exception as x:
            LOG.error(x)
        return {'ignored_brands': {},
                'favorite_brands': {},
                'specially_requested': {}}

    def preference_user(self, message=None) -> dict:
        """
        Returns the user dictionary with name, email
        Equivalent to self.user_config["user"] for non-server use
        """
        try:
            nick = get_message_user(message) if message else None
            if self.server:
                if not message or not nick:
                    LOG.warning("No message given!")
                    return self.user_config['user']
                if message.context.get("nick_profiles"):
                    return message.context["nick_profiles"][nick]["user"]
                else:
                    LOG.error(f"Unable to get user settings! message={message.data}")
            else:
                return self.user_config['user']
        except Exception as x:
            LOG.error(x)
        return {'first_name': '',
                'middle_name': '',
                'last_name': '',
                'preferred_name': '',
                'full_name': '',
                'dob': 'YYYY/MM/DD',
                'age': '',
                'email': '',
                'username': '',
                'password': '',
                'picture': '',
                'about': '',
                'phone': '',
                'email_verified': False,
                'phone_verified': False
                }

    def preference_location(self, message=None) -> dict:
        """
        Get the JSON data structure holding location information.
        Equivalent to self.user_config["location"] for non-server use
        """
        try:
            nick = get_message_user(message) if message else None
            if self.server:
                if not message or not nick:
                    LOG.warning("No message given!")
                    return self.user_config['location']
                if message.context.get("nick_profiles"):
                    return message.context["nick_profiles"][nick]["location"]
                else:
                    LOG.error(f"Unable to get user settings! message={message.data}")
            else:
                return self.user_config['location']
        except Exception as x:
            LOG.error(x)
        return {'lat': 47.4799078,
                'lng': -122.2034496,
                'city': 'Renton',
                'state': 'Washington',
                'country': 'USA',
                'tz': 'America/Los_Angeles',
                'utc': -8.0
                }

    def preference_unit(self, message=None) -> dict:
        """
        Returns the units dictionary that contains time, date, measure formatting preferences
        Equivalent to self.user_config["units"] for non-server use
        """
        try:
            nick = get_message_user(message) if message else None
            if self.server:
                if not message or not nick:
                    LOG.warning("No message given!")
                    return self.user_config['units']

                if message.context.get("nick_profiles"):
                    return message.context["nick_profiles"][nick]["units"]
                else:
                    LOG.error(f"Unable to get user settings! message={message.data}")
            else:
                return self.user_config['units']
        except Exception as x:
            LOG.error(x)
        return {'time': 12,
                'date': 'MDY',
                'measure': 'imperial'
                }

    def preference_speech(self, message=None) -> dict:
        """
        Returns the speech dictionary that contains language and spoken response preferences
        Equivalent to self.user_config["speech"] for non-server use
        """
        try:
            nick = get_message_user(message) if message else None
            if self.server:
                if not message or not nick:
                    LOG.warning("No message given!")
                    return self.user_config['speech']

                if message.context.get("nick_profiles"):
                    return message.context["nick_profiles"][nick]["speech"]
                else:
                    LOG.error(f"Unable to get user settings! message={message.data}")
            else:
                return self.user_config['speech']
        except Exception as x:
            LOG.error(x)
        return {'stt_language': 'en',
                'stt_region': 'US',
                'alt_languages': ['en'],
                'tts_language': "en-us",
                'tts_gender': 'female',
                'neon_voice': 'Joanna',
                'secondary_tts_language': '',
                'secondary_tts_gender': '',
                'secondary_neon_voice': '',
                'speed_multiplier': 1.0,
                'synonyms': {}
                }

    def preference_skill(self, message=None) -> dict:
        """
        Returns the skill settings configuration
        Equivalent to self.settings for non-server
        :param message: Message associated with request
        :return: dict of skill preferences
        """
        nick = get_message_user(message) if message else None
        if self.server and nick:
            try:
                skill = self.skill_id
                LOG.info(f"Get server prefs for skill={skill}")
                user_overrides = message.context["nick_profiles"][nick]["skills"].get(self.skill_id, dict())
                LOG.debug(user_overrides)
                merged_settings = {**self.settings, **user_overrides}
                if user_overrides.keys() != merged_settings.keys():
                    LOG.info(f"New settings keys: user={nick}|skill={self.skill_id}|user={user_overrides}")
                    self.update_skill_settings(merged_settings, message)
                return merged_settings
            except Exception as e:
                LOG.error(e)
        return self.settings

    def build_user_dict(self, message=None) -> dict:
        """
        Builds a merged dictionary containing all user preferences in a single-level dictionary.
        Used to build a dictionary for server profile updates
        :param message: Message associate with request
        """
        merged_dict = {**self.preference_speech(message), **self.preference_user(message),
                       **self.preference_brands(message), **self.preference_location(message),
                       **self.preference_unit(message)}
        for key, value in merged_dict.items():
            if value == "":
                merged_dict[key] = -1
        return merged_dict

    def build_combined_skill_object(self, message=None) -> list:
        user = self.get_utterance_user(message)
        skill_dict = message.context["nick_profiles"][user]["skills"]
        skill_list = list(skill_dict.values())
        return skill_list

    def update_profile(self, new_preferences: dict, message: Message = None):
        """
        Updates a user profile with the passed new_preferences
        :param new_preferences: dict of updated preference values. Should follow {section: {key: val}} format
        :param message: Message associated with request
        """
        if self.server:
            nick = get_message_user(message) if message else None
            new_skills_prefs = new_preferences.pop("skills")
            old_skills_prefs = message.context["nick_profiles"][nick]["skills"]
            combined_skill_prefs = {**old_skills_prefs, **new_skills_prefs}
            combined_changes = {k: v for dic in new_preferences.values() for k, v in dic.items()}
            if new_skills_prefs:
                combined_changes["skill_settings"] = json.dumps(list(combined_skill_prefs.values()))
                new_preferences["skills"] = combined_skill_prefs
                LOG.debug(f"combined_skill_prefs={combined_skill_prefs}")
            combined_changes["username"] = nick
            self.socket_emit_to_server("update profile", ["skill", combined_changes,
                                                          message.context["klat_data"]["request_id"]])
            self.bus.emit(Message("neon.remove_cache_entry", {"nick": nick}))
            old_preferences = message.context["nick_profiles"][nick]
            message.context["nick_profiles"][nick] = {**old_preferences, **new_preferences}
        else:
            for section, settings in new_preferences:
                # section in user, brands, units, etc.
                for key, val in settings:
                    self.user_config.update_yaml_file(section, key, val, True)

            self.user_config.update_yaml_file(final=True)
            modified = ["ngi_user_info"]
            self.bus.emit(Message('check.yml.updates',
                                  {"modified": modified},
                                  {"origin": self.skill_id}))

    def update_skill_settings(self, new_preferences: dict, message: Message = None, skill_global=False):
        """
        Updates skill settings with the passed new_preferences
        :param new_preferences: dict of updated preference values. {key: val}
        :param message: Message associated with request
        :param skill_global: Boolean to indicate these are global/non-user-specific variables
        """
        LOG.debug(f"Update skill settings with new: {new_preferences}")
        if self.server and not skill_global:
            new_preferences["skill_id"] = self.skill_id
            self.update_profile({"skills": {self.skill_id: new_preferences}}, message)
        else:
            for key, val in new_preferences.items():
                self.settings[key] = val
                self.ngi_settings.update_yaml_file(key, value=val, multiple=True)
            self.ngi_settings.update_yaml_file(final=True)

    def build_message(self, kind, utt, message, speaker=None):
        """
        Build a message for user input or neon response
        :param kind: "neon speak" or "execute"
        :param utt: string to emit
        :param message: incoming message object
        :param speaker: speaker data dictionary
        :return: Message object
        """
        LOG.debug(speaker)

        default_speech = self.preference_speech(message)
        # Override user preference for all script responses
        if not speaker:
            speaker = {"name": "Neon",
                       "language": default_speech["tts_language"],
                       "gender": default_speech["tts_gender"],
                       "voice": default_speech["neon_voice"],
                       "override_user": True}
        elif speaker and speaker.get("language"):
            speaker["override_user"] = True
        else:
            speaker = None

        LOG.debug(f"data={message.data}")
        # LOG.debug(f"context={message.context}")

        emit_response = False
        if kind == "skill_data":
            emit_response = True
            kind = "execute"

        try:
            if kind == "execute":
                # This is picked up in the intent handler
                return message.reply("skills:execute.utterance", {
                    "utterances": [utt.lower()],
                    "lang": message.data.get("lang", "en-US"),
                    "session": None,
                    "ident": None,
                    "speaker": speaker
                }, {
                    "neon_should_respond": True,
                    "cc_data": {"request": utt,
                                "emit_response": emit_response,
                                "execute_from_script": True
                                }
                })
            elif kind == "neon speak":
                added_context = {"cc_data": message.context.get("cc_data", {})}
                added_context["cc_data"]["request"] = utt

                return message.reply("speak", {"lang": message.data.get("lang", "en-US"),
                                               "speaker": speaker
                                               }, added_context)
        except Exception as x:
            LOG.error(x)

    def mobile_skill_intent(self, action: str, arguments: dict, message: Message):
        """
        Handle a mobile skill intent response
        :param action: Name of action or event for mobile device to handle
        :param arguments: dict of key/value arguments to pass with action
        :param message: Message associated with request
        """
        fmt_args = ""
        for key, value in arguments:
            fmt_args += f"&{key}={value}"
        if self.server:
            emit_data = [action, fmt_args, message.context["klat_data"]["request_id"]]
            self.bus.emit(Message("css.emit", {"event": "mobile skill intent", "data": emit_data}))
        else:
            LOG.warning("Mobile intents are not supported on this device yet.")

    def socket_emit_to_server(self, event: str, data: list):
        LOG.debug(f"Emit event={event}, data={data}")
        self.bus.emit(Message("css.emit", {"event": event, "data": data}))

    def send_with_audio(self, text_shout, audio_file, message, lang="en-us", private=False, speaker=None):
        """
        Sends a Neon response with the passed text phrase and audio file
        :param text_shout: (str) Text to shout
        :param audio_file: (str) Full path to an arbitrary audio file to attach to shout; must be readable/accessible
        :param message: Message associated with request
        :param lang: (str) Language of wav_file
        :param private: (bool) Whether or not shout is private to the user
        :param speaker: (dict) Message sender data
        """
        # from shutil import copyfile
        if not speaker:
            speaker = {"name": "Neon", "language": None, "gender": None, "voice": None}

        # Play this back regardless of user prefs
        speaker["override_user"] = True

        # Either gender should be fine
        responses = {lang: {"sentence": text_shout,
                            "male": audio_file,
                            "female": audio_file}}
        message.context["private"] = private
        LOG.info(f"sending klat.response with responses={responses} | speaker={speaker}")
        self.bus.emit(message.forward("klat.response", {"responses": responses, "speaker": speaker}))

    def neon_must_respond(self, message):
        """
        Checks if Neon must respond to an utterance (i.e. a server request)
        @param message:
        @return:
        """
        if self.server:
            title = message.context.get("klat_data", {}).get("title", "")
            LOG.debug(message.data.get("utterance"))
            if message.data.get("utterance").startswith("Welcome to your private conversation"):
                return False
            if title.startswith("!PRIVATE:"):
                if ',' in title:
                    users = title.split(':')[1].split(',')
                    for idx, val in enumerate(users):
                        users[idx] = val.strip()
                    if len(users) == 2 and "Neon" in users:
                        # Private with Neon
                        # LOG.debug("DM: Private Conversation with Neon")
                        return True
                    elif message.data.get("utterance").lower().startsWith("neon"):
                        # Message starts with "neon", must respond
                        return True
                else:
                    # Solo Private
                    return True
        return False

    def voc_match(self, utt, voc_filename, lang=None, exact=False):
        # TODO: Handles bug to be addressed in: https://github.com/OpenVoiceOS/ovos_utils/issues/73
        try:
            return super().voc_match(utt, voc_filename, lang, exact)
        except FileNotFoundError:
            LOG.warning(f"`{voc_filename}` not found, checking in neon_core")
            from mycroft.skills.skill_data import read_vocab_file
            from itertools import chain
            import re
        lang = lang or self.lang
        voc = os.path.join(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))), "neon_core",
                           "res", "text", lang, f"{voc_filename}.voc")
        if not os.path.exists(voc):
            raise FileNotFoundError(voc)
        vocab = read_vocab_file(voc)
        cache_key = lang + voc_filename
        self.voc_match_cache[cache_key] = list(chain(*vocab))
        if utt:
            if exact:
                # Check for exact match
                return any(i.strip() == utt
                           for i in self.voc_match_cache[cache_key])
            else:
                # Check for matches against complete words
                return any([re.match(r'.*\b' + i + r'\b.*', utt)
                            for i in self.voc_match_cache[cache_key]])
        else:
            return False

    def neon_in_request(self, message):
        """
        Checks if the utterance is intended for Neon. Server utilizes current conversation, otherwise wake-word status
        and message "Neon" parameter used
        """
        if message.context.get("neon_should_respond", False):
            return True
        elif message.data.get("Neon") or message.data.get("neon"):
            return True
        elif not self.server and self.local_config.get("interface", {}).get("wake_word_enabled", True):
            return True
        elif self.server and message.context.get("klat_data", {}).get("title").startswith("!PRIVATE"):
            return True
        else:
            try:
                voc_match = self.voc_match(message.data.get("utterance"), "neon")
                if voc_match:
                    return True
            except FileNotFoundError:
                LOG.error(f"No neon vocab found!")
                if "neon" in message.data.get("utterance").lower():
                    return True
            LOG.debug("No Neon")
            return False

    def show_settings_gui(self):
        """
        Function to update and
        :return:
        """
        try:
            # TODO: Conditionalize register, only needs to happen once but only after skill init DM
            self.gui.register_settings()
            self.gui.show_settings()
        except Exception as e:
            LOG.error(e)

    def check_yes_no_response(self, message):
        """
        Used in converse methods to check if a response confirms or declines an action. Differs from ask_yesno in that
        this does not assume input will be spoken with wake words enabled
        :param message: incoming message object to evaluate
        :return: False if declined, numbers if confirmed numerically, True if confirmed with no numbers
        """
        utterance = message.data.get("utterances")[0]
        if self.voc_match(utterance, "no"):
            LOG.info("User Declined")
            return False
        elif self.voc_match(utterance, "yes"):
            LOG.info("User Accepted")
            numbers = [str(s) for s in utterance.split() if s.isdigit()]
            if numbers and len(numbers) > 0:
                confirmation = "".join(numbers)
                LOG.info(f"Got confirmation: {confirmation}")
                return confirmation
            return True
        else:
            LOG.debug("User response not valid")
            return -1

    def report_metric(self, name, data):
        """Report a skill metric to the Mycroft servers.

        Arguments:
            name (str): Name of metric. Must use only letters and hyphens
            data (dict): JSON dictionary to report. Must be valid JSON
        """
        combinded = deepcopy(data)
        combinded["name"] = name
        self.bus.emit(Message("neon.metric", combinded))

    def send_email(self, title, body, message=None, email_addr=None, attachments=None):
        """
        Send an email to the registered user's email.
        Email address priority: email_addr, user prefs from message, fallback to DeviceApi for Mycroft method

        Arguments:
            title (str): Title of email
            body  (str): HTML body of email. This supports
                         simple HTML like bold and italics
            email_addr (str): Optional email address to use
            attachments (dict): Optional dict of file names to Base64 encoded files
            message (Message): Optional message to get email from
        """
        if not email_addr and message:
            email_addr = self.preference_user(message).get("email")

        if email_addr:
            LOG.info("Send email via Neon Server")
            try:
                LOG.debug(f"body={body}")
                self.bus.emit(Message("neon.send_email", {"title": title, "email": email_addr, "body": body,
                                                          "attachments": attachments}))
            except Exception as e:
                LOG.error(e)
        else:
            super().send_email(title, body)

    def make_active(self, duration_minutes=5):
        """Bump skill to active_skill list in intent_service.

        This enables converse method to be called even without skill being
        used in last 5 minutes.
        :param duration_minutes: duration in minutes for skill to remain active (-1 for infinite)
        """
        self.bus.emit(Message("active_skill_request",
                              {"skill_id": self.skill_id,
                               "timeout": duration_minutes}))

    def register_decorated(self):
        """
        Accessor method
        """
        self._register_decorated()

    def speak(self, utterance, expect_response=False, wait=False, meta=None, message=None, private=False, speaker=None):
        """
        Speak a sentence.
        Arguments:
            utterance (str):        sentence mycroft should speak
            expect_response (bool): set to True if Mycroft should listen for a response immediately after
                                    speaking the utterance.
            message (Message):      message associated with the input that this speak is associated with
            private (bool):         flag to indicate this message contains data that is private to the requesting user
            speaker (dict):         dict containing language or voice data to override user preference values
            wait (bool):            set to True to block while the text is being spoken.
            meta:                   Information of what built the sentence.
        """
        # registers the skill as being active
        meta = meta or {}
        meta['skill'] = self.name
        self.enclosure.register(self.name)
        if utterance:
            if not message:
                # Find the associated message
                LOG.debug('message is None.')
                message = dig_for_message()
                if not message:
                    message = Message("speak")
            if not speaker:
                speaker = message.data.get("speaker", None)

            nick = get_message_user(message)

            if private and self.server:
                LOG.debug("Private Message")
                title = message.context["klat_data"]["title"]
                need_at_sign = True
                if title.startswith("!PRIVATE"):
                    users = title.split(':')[1].split(',')
                    for idx, val in enumerate(users):
                        users[idx] = val.strip()
                    if len(users) == 2 and "Neon" in users:
                        need_at_sign = False
                    elif len(users) == 1:
                        need_at_sign = False
                    elif nick.startswith("guest"):
                        need_at_sign = False
                if need_at_sign:
                    LOG.debug("Send message to private cid!")
                    utterance = f"@{nick} {utterance}"

            data = {"utterance": utterance,
                    "expect_response": expect_response,
                    "meta": meta,
                    "speaker": speaker}

            if message.context.get("cc_data", {}).get("emit_response"):
                msg_to_emit = message.reply("skills:execute.response", data)
            else:
                message.context.get("timing", {})["speech_start"] = time.time()
                msg_to_emit = message.reply("speak", data, message.context)
                LOG.debug(f"Skill speak! {data}")

            LOG.debug(msg_to_emit.msg_type)
            self.bus.emit(msg_to_emit)
        else:
            LOG.warning("Null utterance passed to speak")
            LOG.warning(f"{self.name} | message={message}")

        if wait:
            wait_while_speaking()

    def speak_dialog(self, key, data=None, expect_response=False, wait=False,
                     message=None, private=False, speaker=None):
        """ Speak a random sentence from a dialog file.

        Arguments:
            :param key: dialog file key (e.g. "hello" to speak from the file "locale/en-us/hello.dialog")
            :param data: information used to populate key
            :param expect_response: set to True if Mycroft should listen for a response immediately after speaking.
            :param wait: set to True to block while the text is being spoken.
            :param speaker: optional dict of speaker info to use
            :param private: private flag (server use only)
            :param message: associated message from request
        """
        data = data or {}
        LOG.debug(f"data={data}")
        self.speak(self.dialog_renderer.render(key, data),  # TODO: Pass index here to use non-random responses DM
                   expect_response, message=message, private=private,
                   speaker=speaker, wait=wait, meta={'dialog': key, 'data': data})

    def schedule_event(self, handler, when, data=None, name=None, context=None):
        # TODO: should 'when' already be a datetime? DM
        if isinstance(when, int) or isinstance(when, float):
            from datetime import datetime as dt, timedelta
            when = to_system_time(dt.now(self.sys_tz)) + timedelta(seconds=when)
            LOG.info(f"Made a datetime: {when}")
        super().schedule_event(handler, when, data, name, context)

    def request_check_timeout(self, time_wait, intent_to_check):
        LOG.info("request received")
        LOG.info(time_wait)
        LOG.info(len(intent_to_check))
        try:
            if isinstance(intent_to_check, str):
                intent_to_check = [intent_to_check]

            for intent in intent_to_check:
                data = {'time_out': time_wait,
                        'intent_to_check': f"{self.skill_id}:{intent}"}
                LOG.debug(f"DM: Set Timeout: {data}")
                self.bus.emit(Message("set_timeout", data))
        except Exception as x:
            LOG.error(x)

    def await_confirmation(self, user, actions, timeout=None):
        """
        Used to add an action for which to await a response (note: this will disable skill reload when called and enable
        on timeout)
        :param user: username ("local" for non-server)
        :param actions: string action name (or list of action names) we are confirming,
                       handled in skill's converse method
        :param timeout: duration to wait in seconds before removing the action from the list
        """
        from datetime import datetime as dt, timedelta
        self.reload_skill = False
        if isinstance(actions, str):
            actions = [actions]
        self.actions_to_confirm[user] = actions
        if not timeout:
            timeout = self.default_intent_timeout
        expiration = dt.now(self.sys_tz) + timedelta(seconds=timeout)

        self.cancel_scheduled_event(user)
        time.sleep(1)
        self.schedule_event(self._confirmation_timeout, to_system_time(expiration), data={"user": user,
                                                                                          "action": actions},
                            name=user)
        LOG.debug(f"Scheduled {user}")

    def _confirmation_timeout(self, message):
        user = message.data.get("user", "local")
        try:
            if user in self.actions_to_confirm.keys():
                removed = self.actions_to_confirm.pop(user)
                LOG.info(f"confirmation timed out ({time.time()}): {removed}")
        except Exception as e:
            # Catches if the item was already popped
            LOG.error(e)
        if len(self.actions_to_confirm.keys()) == 0:
            self.reload_skill = True

    def clear_gui_timeout(self, timeout_seconds=60):
        """
        Called by a skill to clear its gui display after the specified timeout
        :param timeout_seconds: seconds to wait before clearing gui display
        :return:
        """
        from datetime import datetime as dt, timedelta
        expiration = dt.now(self.sys_tz) + timedelta(seconds=timeout_seconds)
        self.schedule_event(self._clear_gui_timeout, to_system_time(expiration))

    def _clear_gui_timeout(self):
        """
        Handler for clear_gui_timeout function
        :return:
        """
        LOG.info("Reset GUI!")
        self.gui.clear()

    def clear_signals(self, prefix):
        """
        Clears all signals that begin with the passed prefix. Used with skill prefix for a skill to clear any signals it
        may have set
        :param prefix: (str) prefix to match
        """
        os.makedirs(f"{self.configuration_available['dirVars']['ipcDir']}/signal", exist_ok=True)
        for signal in os.listdir(self.configuration_available['dirVars']['ipcDir'] + '/signal'):
            if str(signal).startswith(prefix) or f"_{prefix}_" in str(signal):
                # LOG.info('Removing ' + str(signal))
                # os.remove(self.configuration_available['dirVars']['ipcDir'] + '/signal/' + signal)
                self.check_for_signal(signal)

    def update_cached_data(self, filename, new_element):
        """
        Updates cache file of skill responses to translated responses when non-english responses are requested.
        :param filename: (str) filename of cache object to update (relative to cacheDir)
        :param new_element: (any) object to cache at passed location
        """
        with open(os.path.join(self.cache_loc, filename), 'wb+') as file_to_update:
            pickle.dump(new_element, file_to_update, protocol=pickle.HIGHEST_PROTOCOL)

    def get_cached_data(self, filename, file_loc=None):
        """
        Retrieves cache data from a file created/updated with update_cached_data
        :param filename: (str) filename of cache object to update
        :param file_loc: (str) path to directory containing filename (defaults to cache dir)
        :return: (dict) cache data
        """
        if not file_loc:
            file_loc = self.cache_loc
        cached_location = os.path.join(file_loc, filename)
        if pathlib.Path(cached_location).exists():
            with open(cached_location, 'rb') as file:
                return pickle.load(file)
        else:
            return {}

    def get_utterance_user(self, message: Optional[Message]) -> str:
        """
        Gets the user associated with the given message. Returns default 'local' or 'server' if no user specified.
        Args:
            message: Message associated with request

        Returns:
            Username associated with the message or a default value of 'local' or 'server'.
        """
        if self.server:
            default_user = "server"
        else:
            default_user = self.preference_user(message).get("username", "local")
        if not message:
            return default_user

        try:
            return get_message_user(message) or default_user
        except Exception as e:
            LOG.error(e)
            # TODO: Depreciate this and fix underlying error DM
            return default_user

    @staticmethod
    def newest_file_in_dir(path, ext=None):
        LOG.warning("This method is depreciated, use file_utils.get_most_recent_file_in_dir() directly")
        return get_most_recent_file_in_dir(path, ext)

    @staticmethod
    def request_from_mobile(message):
        LOG.warning("This method is depreciated, use message_utils.request_from_mobile() directly")
        return request_from_mobile(message)

    @staticmethod
    def to_system_time(dt):
        LOG.warning("This method is depreciated, use location_utils.to_system_time() directly")
        return to_system_time(dt)

    def decorate_api_call_use_lru(self, func):
        """
        Decorate the API-call function to use LRUcache.
        NOTE: the wrapper adds an additional argument, so decorated functions MUST be called with it!

        from wikipedia_for_humans import summary
        summary = decorate_api_call_use_lru(summary)
        result = summary(lru_query='neon', query='neon', lang='en')

        Args:
            func: the function to be decorated
        Returns: decorated function
        """
        @wraps(func)
        def wrapper(lru_query: str, *args, **kwargs):
            # TODO might use an abstract method for cached API call to define a signature
            result = self.lru_cache.get(lru_query)
            if not result:
                result = func(*args, **kwargs)
                self.lru_cache.put(key=lru_query, value=result)
            return result
        return wrapper

    def _write_cache_on_disk(self):
        """
        Write the cache on disk, reset the cache and reschedule the event.
        This handler is enabled by scheduling an event in NeonSkill.initialize().
        Returns:
        """
        filename = f"lru_{self.skill_id}"
        data_load = self.lru_cache.jsonify()
        self.update_cached_data(filename=filename, new_element=data_load)
        self.lru_cache.clear()
        self.schedule_event(self._write_cache_on_disk, CACHE_TIME_OFFSET, name="neon.load_cache_on_disk")
        return



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from neon_utils.skill_override_functions import *
from neon_utils.variables_override import *
from neon_utils.logger import LOG

from ovos_utils.gui import is_gui_installed

SKILL = None
TYPE = None


def skill_needs_patching(skill):
    """
    Determines if the passed skill is running under a non-Neon core and needs to be patched for compatibility
    :param skill: MycroftSkill object to test
    :return: True if skill needs to be patched
    """
    LOG.warning(f"This method is depreciated. Please update your skill to extend NeonSkill instead.")
    return not hasattr(skill, "neon_core")


def stub_missing_parameters(skill):
    global SKILL
    global TYPE

    LOG.warning(f"This method is depreciated. Please update your skill to extend NeonSkill instead.")

    SKILL = skill
    TYPE = type(skill)
    LOG.debug(SKILL)
    LOG.debug(TYPE)

    skill.actions_to_confirm = dict()
    skill.default_intent_timeout = None
    skill.server = False
    skill.gui_enabled = is_gui_installed()  # This isn't a check for running, just available DM

    skill.neon_in_request = neon_in_request
    skill.neon_must_respond = neon_must_respond
    skill.request_from_mobile = request_from_mobile
    skill.speak_dialog = speak_dialog
    skill.speak = speak
    skill.create_signal = create_signal
    skill.check_for_signal = check_for_signal
    skill.clear_signals = clear_signals

    skill.preference_brands = preference_brands
    skill.preference_user = preference_user
    skill.preference_location = preference_location
    skill.preference_unit = preference_unit
    skill.preference_speech = preference_speech
    skill.preference_skill = preference_skill
    skill.build_user_dict = build_user_dict
    skill.request_check_timeout = request_check_timeout
    skill.get_utterance_user = get_utterance_user
    skill.update_skill_settings = update_skill_settings

    skill.neon_core = False
    skill.configuration_available = configuration_available

    try:
        # TODO: This should really be global to match Neon.. Maybe /opt/mycroft? DM
        skill.cache_loc = os.path.join(skill.__location__, "cache")
    except Exception as e:
        LOG.error(e)
        skill.cache_loc = os.path.join("tmp", "mycroft", "cache")

    skill.get_cached_data = get_cached_data
    skill.update_cached_data = update_cached_data
    skill.build_message = build_message

    skill.speak = speak
    skill.wait_while_speaking = wait_while_speaking
    skill.is_speaking = is_speaking
    skill.to_system_time = to_system_time

    skill.check_yes_no_response = check_yes_no_response



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import nltk
import re
import os

from neon_utils.logger import LOG


def clean_quotes(raw_utt: str) -> str:
    """
    Method for stripping quotes from fully quoted strings in different languages
    :param raw_utt: Input string to be cleaned
    :return: string with all paired quote characters removed
    """
    if not raw_utt:
        raise ValueError("Expected a string and got None")
    if not isinstance(raw_utt, str):
        raise TypeError(f"{raw_utt} is not a string!")
    chars_to_remove = ['“', '"', '«', u'\u201d', u'\u00bb', u'\u201e', '「', '」', u'u\xa0', u'\u00a0']
    raw_utt = raw_utt.strip()
    utt = raw_utt
    trailing_punctuation = False
    if utt.endswith("."):
        trailing_punctuation = True
        utt = utt.rstrip(".")
    quotes_cleaned = False
    try:
        # Checks if utterance starts AND ends with some form of quotations and removes them accordingly
        while (utt.startswith('“') or utt.startswith(u'\u201d') or utt.startswith('"') or utt.startswith('«')
               or utt.startswith(u'\u00bb') or utt.startswith(u'\u201e') or utt.startswith('「') or
               utt.startswith(u'u\xa0') or utt.startswith(u'\u00a0')) and \
                (utt.endswith('“') or utt.endswith(u'\u201d') or utt.endswith('"') or utt.endswith(u'\u00bb') or
                 utt.endswith(u'\u201e') or utt.endswith('」') or utt.endswith(u'u\xa0') or
                 utt.endswith(u'\u00a0') or utt.endswith('»')):
            quotes_cleaned = True
            removed_left, removed_right = False, False
            for c in chars_to_remove:
                if not removed_left and utt.startswith(c):
                    utt = utt[1:]
                    removed_left = True
                if not removed_right and utt.endswith(c):
                    utt = utt[:-1]
                    removed_right = True
        if quotes_cleaned:
            if trailing_punctuation:
                return f"{utt}."
            return utt
        else:
            return raw_utt
    except Exception as x:
        LOG.error(x)
        return raw_utt


def clean_filename(raw_name: str, to_lowercase: bool = False) -> str:
    """
    Cleans a filename of any invalid characters
    :param raw_name: input file basename to clean
    :param to_lowercase: cast string to lowercase
    :return: cleaned file basename
    """
    if not raw_name:
        raise ValueError
    invalid_chars = ('/', '\\', '*', '~', ':', '"', '<', '>', '|', '?')
    name = raw_name
    for char in invalid_chars:
        name = name.replace(char, "_")
    if to_lowercase:
        name = name.lower()
    return name


def clean_transcription(raw_string: str) -> str:
    """
    Cleans up input transcriptions to replace any special characters with text
    :param raw_string: Input string to be cleaned
    :return: Cleaned string of alphas
    """
    if not raw_string:
        raise ValueError
    parsed = raw_string.lower().replace('%', ' percent').replace('.', ' ').replace('?', '').replace('-', ' ').strip()
    # TODO: Cleanup to alpha string (replace all chars)
    return parsed


def get_phonemes(phrase: str) -> str:
    """
    Gets phonemes for the requested phrase
    :param phrase: String phrase for which to get phonemes
    :return: ARPAbet phonetic representation (https://en.wikipedia.org/wiki/ARPABET)
    """
    download_path = os.path.expanduser("~/.local/share/neon")
    if not os.path.isdir(download_path):
        os.makedirs(download_path)
    nltk.download('cmudict', download_dir=download_path)
    nltk.data.path.append(download_path)

    output = ''
    for word in phrase.split():
        phoenemes = nltk.corpus.cmudict.dict()[word.lower()][0]
        for phoeneme in phoenemes:
            output += str(re.sub('[0-9]', '', phoeneme) + ' ')
        output += '. '
    return output.rstrip()


def format_speak_tags(sentence: str, include_tags: bool = True) -> str:
    """
    Cleans up SSML tags for speech synthesis and ensures the phrase is wrapped in 'speak' tags and any excluded text is
    removed.
    Args:
        sentence: Input sentence to be spoken
        include_tags: Flag to include <speak> tags in returned string
    Returns:
        Cleaned sentence to pass to TTS
    """
    # Wrap sentence in speak tag if no tags present
    if "<speak>" not in sentence and "</speak>" not in sentence:
        to_speak = f"<speak>{sentence}</speak>"
    # Assume speak starts at the beginning of the sentence if a closing speak tag is found
    elif "<speak>" not in sentence:
        to_speak = f"<speak>{sentence}"
    # Assume speak ends at the end of the sentence if an opening speak tag is found
    elif "</speak>" not in sentence:
        to_speak = f"{sentence}</speak>"
    else:
        to_speak = sentence

    # Trim text outside of speak tags
    if not to_speak.startswith("<speak>"):
        to_speak = f"<speak>{to_speak.split('<speak>', 1)[1]}"

    if not to_speak.endswith("</speak>"):
        to_speak = f"{to_speak.split('</speak>', 1)[0]}</speak>"

    if to_speak == "<speak></speak>":
        return ""

    if include_tags:
        return to_speak
    else:
        return to_speak.lstrip("<speak>").rstrip("</speak>")


def normalize_string_to_speak(to_speak: str) -> str:
    """
    Normalizes spoken strings for TTS engines to handle
    :param to_speak: String to speak
    :return: string with any invalid characters removed and punctuation added
    """
    if not to_speak:
        raise ValueError("Expected a string and got None")
    if not isinstance(to_speak, str):
        raise TypeError(f"{to_speak} is not a string!")

    valid_punctuation = ['.', '?', '!']
    if any(to_speak.endswith(x) for x in valid_punctuation):
        return to_speak
    return f"{to_speak}."



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import logging
import re
import json
import os
import sys
import shutil
import sysconfig

from copy import deepcopy
from os.path import *
from collections.abc import MutableMapping
from contextlib import suppress
from filelock import FileLock
from glob import glob
from ovos_utils.json_helper import load_commented_json
from ovos_utils.configuration import read_mycroft_config, LocalConf
from ruamel.yaml import YAML
from typing import Optional
from neon_utils import LOG
from neon_utils.authentication_utils import find_neon_git_token, populate_github_token_config

logging.getLogger("filelock").setLevel(logging.WARNING)


class NGIConfig:
    configuration_list = dict()

    def __init__(self, name, path=None, force_reload: bool = False):
        self.name = name
        self.path = path or get_config_dir()
        self.parser = YAML()
        lock_filename = join(self.path, f".{self.name}.lock")
        self.lock = FileLock(lock_filename, timeout=10)
        self._pending_write = False
        self._content = dict()
        self._loaded = os.path.getmtime(self.file_path)
        if not force_reload and self.__repr__() in NGIConfig.configuration_list:
            cache = NGIConfig.configuration_list[self.__repr__()]
            cache.check_reload()
            self._content = cache.content
        else:
            self._content = self._load_yaml_file()
            NGIConfig.configuration_list[self.__repr__()] = self
        self._disk_content_hash = hash(repr(self._content))

    @property
    def requires_reload(self):
        return self._loaded != os.path.getmtime(self.file_path)

    def check_reload(self):
        if self._loaded != os.path.getmtime(self.file_path):
            self.check_for_updates()

    def write_changes(self):
        if self._pending_write or self._disk_content_hash != hash(repr(self._content)):
            self._write_yaml_file()

    def populate(self, content, check_existing=False):
        if not check_existing:
            self.__add__(content)
            return
        old_content = deepcopy(self._content)
        self._content = dict_merge(content, self._content)  # to_change, one_with_all_keys
        if old_content == self._content:
            LOG.warning(f"Update called with no change: {self.file_path}")
            return
        self._write_yaml_file()

    def remove_key(self, *key):
        for item in key:
            self.__sub__(item)

    def make_equal_by_keys(self, other: MutableMapping, recursive: bool = True, depth: int = 1):
        """
        Adds and removes keys from this config such that it has the same keys as 'other'. Configuration values are
        preserved with any added keys using default values from 'other'.
        Args:
            other: dict of keys and default values this configuration should have
            recursive: flag to indicate configuration may be merged recursively
            depth: int depth to recurse (0 includes top-level keys only)
        """
        old_content = deepcopy(self._content)
        if not recursive:
            depth = 0
        self._content = dict_make_equal_keys(self._content, other, depth)
        if old_content == self._content:
            return
        self._write_yaml_file()

    def update_keys(self, other):
        """
        Adds keys to this config such that it has all keys in 'other'. Configuration values are
        preserved with any added keys using default values from 'other'.
        Args:
            other: dict of keys and default values this should be added to this configuration
        """
        old_content = deepcopy(self._content)
        self._content = dict_update_keys(self._content, other)  # to_change, one_with_all_keys
        if old_content == self._content:
            LOG.warning(f"Update called with no change: {self.file_path}")
            return

        self._write_yaml_file()

    @property
    def file_path(self):
        """
        Returns the path to the yml file associated with this configuration
        Returns: path to this configuration yml
        """
        file_path = join(self.path, self.name + ".yml")
        if not isfile(file_path):
            create_file(file_path)
            LOG.debug(f"New YAML created: {file_path}")
        return file_path

    def check_for_updates(self) -> dict:
        """
        Reloads updated configuration from disk. Used to reload changes when other instances modify a configuration
        Returns:Updated configuration.content
        """
        new_content = self._load_yaml_file()
        if new_content:
            LOG.debug(f"{self.name} Checked for Updates")
            self._content = new_content
        else:
            LOG.error("new_content is empty!!")
            new_content = self._load_yaml_file()
            if new_content:
                LOG.debug("second attempt success")
                self._content = new_content
            else:
                LOG.error("second attempt failed")
        return self._content

    def update_yaml_file(self, header=None, sub_header=None, value="", multiple=False, final=False):
        """
        Called by class's children to update, create, or initiate a new parameter in the
        specified YAML file. Creates and updates headers, adds or overwrites preference elements,
        associates value to the created or existing field. Recursive if creating a new
        header-preference-value combo.
        :param multiple: true if more than one continuous write is coming
        :param header: string with the new or existing main header
        :param sub_header: new or existing subheader (sublist)
        :param value: any value that should be associated with the headers.
        :param final: true if this is the last change when skip_reload was true
        :return: pre-existing parameter if nothing to update or error if invalid yaml_type.
        """
        # with self.lock.acquire(30):
        self.check_reload()
        before_change = self._content

        LOG.debug(value)
        if header and sub_header:
            try:
                before_change[header][sub_header] = value
            except KeyError:
                before_change[header] = {sub_header: value}
                return
        elif header and not sub_header:
            try:
                before_change[header] = value
            except Exception as x:
                LOG.error(x)
        else:
            LOG.debug("No change needed")
            if not final:
                return

        if not multiple:
            self._write_yaml_file()
        else:
            LOG.debug("More than one change")
            self._pending_write = True

    def export_to_json(self) -> str:
        """
        Export this configuration to a json file
        Returns: path to exported file
        """
        json_filename = os.path.join(self.path, f"{self.name}.json")
        write_to_json(self._content, json_filename)
        return json_filename

    def from_dict(self, pref_dict: dict):
        """
        Constructor to build this configuration object with the passed dict of data
        Args:
            pref_dict: dict to populate configuration with

        Returns: this object

        """
        self._content = pref_dict

        self._write_yaml_file()
        return self

    def from_json(self, json_path: str):
        """
        Constructor to build this configuration object with the passed json file
        Args:
            json_path: Path to json file to populate configuration with

        Returns: this object

        """
        self._content = load_commented_json(json_path)

        self._write_yaml_file()
        return self

    def _load_yaml_file(self) -> dict:
        """
        Loads and parses the YAML file at a given filepath into the Python
        dictionary object.
        :return: dictionary, containing all keys and values from the most current
                 selected YAML.
        """
        try:
            self._loaded = os.path.getmtime(self.file_path)
            with self.lock:
                with open(self.file_path, 'r') as f:
                    return self.parser.load(f) or dict()
        except FileNotFoundError as x:
            LOG.error(f"Configuration file not found error: {x}")
        except Exception as c:
            LOG.error(f"{self.file_path} Configuration file error: {c}")
        return dict()

    def _write_yaml_file(self):
        """
        Overwrites and/or updates the YML at the specified file_path.
        """
        try:
            with self.lock.acquire(30):
                tmp_filename = join(self.path, f".{self.name}.tmp")
                LOG.debug(f"tmp_filename={tmp_filename}")
                shutil.copy2(self.file_path, tmp_filename)
                with open(self.file_path, 'w+') as f:
                    self.parser.dump(self._content, f)
                    LOG.debug(f"YAML updated {self.name}")
                self._loaded = os.path.getmtime(self.file_path)
                self._pending_write = False
                self._disk_content_hash = hash(repr(self._content))
        except FileNotFoundError as x:
            LOG.error(f"Configuration file not found error: {x}")

    @property
    def content(self) -> dict:
        """
        Loads any changes from disk and returns an updated configuration dict
        Returns:
        dict content of this configuration object
        """
        self.check_reload()
        return self._content

    def get(self, *args) -> any:
        """
        Wraps content.get() to provide standard access to an updated configuration like a Python dictionary.
        Args:
            *args: args passed to self.content.get() (key and default value)

        Returns:
        self.content.get(*args)
        """
        return self.content.get(*args)

    def __getitem__(self, item):
        return self.content.get(item)

    def __contains__(self, item):
        return item in self._content

    def __setitem__(self, key, value):
        LOG.info(f"Config changes pending write to disk!")
        self._pending_write = True
        self._content[key] = value

    def __repr__(self):
        return f"NGIConfig('{self.name}')@{self.file_path}"

    def __str__(self):
        return f"{self.file_path}: {json.dumps(self._content, indent=4)}"

    def __add__(self, other):
        # with self.lock.acquire(30):
        if other:
            if not isinstance(other, NGIConfig) and not isinstance(other, MutableMapping):
                raise AttributeError("__add__ expects dict or config object as argument")
            to_update = other
            if isinstance(other, NGIConfig):
                to_update = other._content
            if self._content:
                self._content.update(to_update)
            else:
                self._content = to_update
        else:
            raise TypeError("__add__ expects an argument other than None")
        self._write_yaml_file()

    def __sub__(self, *other):
        # with self.lock.acquire(30):
        if other:
            for element in other:
                if isinstance(element, NGIConfig):
                    to_remove = list(element._content.keys())
                elif isinstance(element, MutableMapping):
                    to_remove = list(element.keys())
                elif isinstance(element, list):
                    to_remove = element
                elif isinstance(element, str):
                    to_remove = [element]
                else:
                    raise AttributeError("__add__ expects dict, list, str, or config object as the argument")

                if self._content:
                    self._content = delete_recursive_dictionary_keys(self._content, to_remove)
                else:
                    raise TypeError("{} config is empty".format(self.name))
        else:
            raise TypeError("__sub__ expects an argument other than None")
        self._write_yaml_file()


def get_config_dir():
    """
    Get a default directory in which to find configuration files
    Returns: Path to configuration or else default
    """
    site = sysconfig.get_paths()['platlib']
    if exists(join(site, 'NGI')):
        return join(site, "NGI")
    for p in [path for path in sys.path if path != ""]:
        if exists(join(p, "NGI")):
            return join(p, "NGI")
        if re.match(".*/lib/python.*/site-packages", p):
            clean_path = "/".join(p.split("/")[0:-4])
            if exists(join(clean_path, "NGI")):
                return join(clean_path, "NGI")
            elif exists(join(clean_path, "mycroft")):
                return clean_path
            elif exists(join(clean_path, ".venv")):
                # LOG.info(f"Saving config to .venv path: {clean_path}")
                return clean_path
    default_path = expanduser("~/.local/share/neon")
    # LOG.info(f"System packaged core found! Using default configuration at {default_path}")
    return default_path


def create_file(filename):
    """ Create the file filename and create any directories needed

        Args:
            filename: Path to the file to be created
    """
    try:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
    except OSError:
        pass
    with open(filename, 'w') as f:
        f.write('')


def delete_recursive_dictionary_keys(dct_to_change: MutableMapping, list_of_keys_to_remove: list) -> MutableMapping:
    """
    Removes the specified keys from the specified dict.
    Args:
        dct_to_change: Dictionary to modify
        list_of_keys_to_remove: List of keys to remove

    Returns: dct_to_change with any specified keys removed

    """
    if not isinstance(dct_to_change, MutableMapping) or not isinstance(list_of_keys_to_remove, list):
        raise AttributeError("delete_recursive_dictionary_keys expects a dict and a list as args")

    for key in list_of_keys_to_remove:
        with suppress(KeyError):
            del dct_to_change[key]
    for value in list(dct_to_change.values()):
        if isinstance(value, MutableMapping):
            delete_recursive_dictionary_keys(value, list_of_keys_to_remove)
    return dct_to_change


def dict_merge(dct_to_change: MutableMapping, merge_dct: MutableMapping) -> MutableMapping:
    """
    Recursively merges two configuration dictionaries and returns the combined object. All keys are returned with values
    from merge_dct overwriting those from dct_to_change.
    Args:
        dct_to_change: dict to append keys and values to
        merge_dct: dict with keys and new values to add to dct_to_change
    Returns: dict of merged preferences
    """
    if not isinstance(dct_to_change, MutableMapping) or not isinstance(merge_dct, MutableMapping):
        raise AttributeError("merge_recursive_dicts expects two dict objects as args")
    for key, value in merge_dct.items():
        if isinstance(dct_to_change.get(key), dict) and isinstance(value, MutableMapping):
            dct_to_change[key] = dict_merge(dct_to_change[key], value)
        else:
            dct_to_change[key] = value
    return dct_to_change


def dict_make_equal_keys(dct_to_change: MutableMapping, keys_dct: MutableMapping,
                         max_depth: int = 1, cur_depth: int = 0) -> MutableMapping:
    """
    Adds and removes keys from dct_to_change such that it has the same keys as keys_dct. Values from dct_to_change are
    preserved with any added keys using default values from keys_dct.
    Args:
        dct_to_change: Dict of user preferences to modify and return
        keys_dct: Dict containing all keys and default values
        max_depth: Int depth to recurse (0-indexed)
        cur_depth: Current depth relative to top-level config (0-indexed)
    Returns: dct_to_change with any keys not in keys_dct removed and any new keys added with default values

    """
    if not isinstance(dct_to_change, MutableMapping) or not isinstance(keys_dct, MutableMapping):
        raise AttributeError("merge_recursive_dicts expects two dict objects as args")
    for key in list(dct_to_change.keys()):
        if isinstance(keys_dct.get(key), dict) and isinstance(dct_to_change[key], MutableMapping):
            if max_depth > cur_depth and key not in ("tts", "stt"):
                dct_to_change[key] = dict_make_equal_keys(dct_to_change[key], keys_dct[key], max_depth, cur_depth + 1)
        elif key not in keys_dct.keys():
            dct_to_change.pop(key)
            LOG.warning(f"Removing '{key}' from dict!")
            # del dct_to_change[key]
    for key, value in keys_dct.items():
        if key not in dct_to_change.keys():
            dct_to_change[key] = value
    return dct_to_change


def dict_update_keys(dct_to_change: MutableMapping, keys_dct: MutableMapping) -> MutableMapping:
    """
    Adds keys to dct_to_change such that all keys in keys_dict exist in dict_to_change. Added keys use default values
    from keys_dict
    Args:
        dct_to_change: Dict of user preferences to modify and return
        keys_dct: Dict containing potentially new keys and default values

    Returns: dct_to_change with any new keys in keys_dict added with default values

    """
    if not isinstance(dct_to_change, MutableMapping) or not isinstance(keys_dct, MutableMapping):
        raise AttributeError("merge_recursive_dicts expects two dict objects as args")
    for key, value in list(keys_dct.items()):
        if isinstance(keys_dct.get(key), dict) and isinstance(value, MutableMapping):
            dct_to_change[key] = dict_update_keys(dct_to_change.get(key, {}), keys_dct[key])
        else:
            if key not in dct_to_change.keys():
                dct_to_change[key] = value
    return dct_to_change


def write_to_json(preference_dict: MutableMapping, output_path: str):
    """
    Writes the specified dictionary to a json file
    Args:
        preference_dict: Dict to write to JSON
        output_path: Output file to write
    """
    if not os.path.exists(output_path):
        create_file(output_path)
    with open(output_path, "w") as out:
        json.dump(preference_dict, out, indent=4)


def get_neon_lang_config() -> dict:
    """
    Get a language config for language utilities
    Returns:
        dict of config params used by Language Detector and Translator modules
    """
    core_config = get_neon_local_config()
    language_config = deepcopy(get_neon_user_config().content.get("speech", {}))
    language_config["internal"] = language_config.get("internal", "en-us")  # TODO: This is core, not user DM
    language_config["user"] = language_config.get("stt_language", "en-us")
    language_config["boost"] = False
    language_config["detection_module"] = core_config.get("stt", {}).get("detection_module")
    language_config["translation_module"] = core_config.get("stt", {}).get("translation_module")

    merged_language = {**_safe_mycroft_config().get("language", {}), **language_config}
    if merged_language.keys() != language_config.keys():
        LOG.warning(f"Keys missing from Neon config! {merged_language.keys()}")

    return merged_language


def get_neon_cli_config() -> dict:
    """
    Get a configuration dict for the neon_cli
    Returns:
        dict of config params used by the neon_cli
    """
    local_config = NGIConfig("ngi_local_conf").content
    wake_words_enabled = local_config.get("interface", {}).get("wake_word_enabled", True)
    try:
        neon_core_version = os.path.basename(glob(local_config['dirVars']['ngiDir'] +
                                                  '/*.release')[0]).split('.release')[0]
    except Exception as e:
        LOG.error(e)
        neon_core_version = "Unknown"
    log_dir = local_config.get("dirVars", {}).get("logsDir", "/var/log/mycroft")
    return {"neon_core_version": neon_core_version,
            "wake_words_enabled": wake_words_enabled,
            "log_dir": log_dir}


def get_neon_tts_config() -> dict:
    """
    Get a configuration dict for TTS
    Returns:
    dict of TTS-related configuration
    """
    return get_neon_local_config()["tts"]


def get_neon_speech_config() -> dict:
    """
    Get a configuration dict for listener. Merge any values from Mycroft config if missing from Neon.
    Returns:
        dict of config params used for listener in neon_speech
    """
    mycroft = _safe_mycroft_config()
    local_config = get_neon_local_config()

    neon_listener_config = deepcopy(local_config.get("listener", {}))
    neon_listener_config["wake_word_enabled"] = local_config["interface"].get("wake_word_enabled", True)
    neon_listener_config["save_utterances"] = local_config["prefFlags"].get("saveAudio", False)
    neon_listener_config["confirm_listening"] = local_config["interface"].get("confirm_listening", True)
    neon_listener_config["record_utterances"] = neon_listener_config["save_utterances"]
    neon_listener_config["record_wake_words"] = neon_listener_config["save_utterances"]
    merged_listener = {**mycroft.get("listener", {}), **neon_listener_config}
    if merged_listener.keys() != neon_listener_config.keys():
        LOG.warning(f"Keys missing from Neon config! {merged_listener.keys()}")

    lang = mycroft.get("language", {}).get("internal", "en-us")  # core_lang

    neon_stt_config = local_config.get("stt", {})
    merged_stt_config = {**mycroft.get("stt", {}), **neon_stt_config}
    if merged_stt_config.keys() != neon_stt_config.keys():
        LOG.warning(f"Keys missing from Neon config! {merged_stt_config.keys()}")

    hotword_config = local_config.get("hotwords") or mycroft.get("hotwords")
    if hotword_config != local_config.get("hotwords"):
        LOG.warning(f"Neon hotword config missing! {hotword_config}")

    neon_audio_parser_config = local_config.get("audio_parsers", {})
    merged_audio_parser_config = {**mycroft.get("audio_parsers", {}), **neon_audio_parser_config}
    if merged_audio_parser_config.keys() != neon_audio_parser_config.keys():
        LOG.warning(f"Keys missing from Neon config! {merged_audio_parser_config.keys()}")

    return {"listener": merged_listener,
            "hotwords": hotword_config,
            "audio_parsers": merged_audio_parser_config,
            "lang": lang,
            "stt": merged_stt_config,
            "metric_upload": local_config["prefFlags"].get("metrics", False),
            "remote_server": local_config.get("remoteVars", {}).get("remoteHost", "64.34.186.120"),
            "data_dir": os.path.expanduser(local_config.get("dirVars", {}).get("rootDir") or "~/.local/share/neon"),
            "keys": {}  # TODO: Read from somewhere DM
            }


def get_neon_bus_config() -> dict:
    """
    Get a configuration dict for the messagebus. Merge any values from Mycroft config if missing from Neon.
    Returns:
        dict of config params used for a messagebus client
    """
    mycroft = _safe_mycroft_config().get("websocket", {})
    neon = get_neon_local_config().get("websocket", {})
    merged = {**mycroft, **neon}
    if merged.keys() != neon.keys():
        LOG.warning(f"Keys missing from Neon config! {merged.keys()}")
    return merged


def get_neon_audio_config() -> dict:
    """
    Get a configuration dict for the audio module. Merge any values from Mycroft config if missing from Neon.
    Returns:
        dict of config params used for the Audio module
    """
    mycroft = _safe_mycroft_config()
    local_config = get_neon_local_config()
    neon_audio = local_config.get("audioService", {})
    merged_audio = {**mycroft.get("Audio", {}), **neon_audio}
    if merged_audio.keys() != neon_audio.keys():
        LOG.warning(f"Keys missing from Neon config! {merged_audio.keys()}")

    return {"Audio": merged_audio,
            "tts": get_neon_tts_config(),
            "language": get_neon_lang_config()}


def get_neon_api_config() -> dict:
    """
    Get a configuration dict for the api module. Merge any values from Mycroft config if missing from Neon.
    Returns:
        dict of config params used for the Mycroft API module
    """
    core_config = get_neon_local_config()
    api_config = deepcopy(core_config.get("api"))
    api_config["metrics"] = core_config["prefFlags"].get("metrics", False)
    mycroft = _safe_mycroft_config().get("server", {})
    merged = {**mycroft, **api_config}
    if merged.keys() != api_config.keys():
        LOG.warning(f"Keys missing from Neon config! {merged.keys()}")
    return merged


def get_neon_skills_config() -> dict:
    """
    Get a configuration dict for the skills module. Merge any values from Mycroft config if missing from Neon.
    Returns:
        dict of config params used for the Mycroft Skills module
    """
    core_config = get_neon_local_config()
    mycroft_config = _safe_mycroft_config()
    neon_skills = deepcopy(core_config.get("skills", {}))
    neon_skills["directory"] = os.path.expanduser(core_config["dirVars"].get("skillsDir"))
    neon_skills["disable_osm"] = neon_skills["skill_manager"] != "osm"
    if not isinstance(neon_skills["auto_update_interval"], float):
        try:
            neon_skills["auto_update_interval"] = float(neon_skills["auto_update_interval"])
        except Exception as e:
            LOG.error(e)
            neon_skills["auto_update_interval"] = 24.0
    if not isinstance(neon_skills["appstore_sync_interval"], float):
        try:
            neon_skills["appstore_sync_interval"] = float(neon_skills["appstore_sync_interval"])
        except Exception as e:
            LOG.error(e)
            neon_skills["appstore_sync_interval"] = 6.0
    neon_skills["update_interval"] = neon_skills["auto_update_interval"]  # Backwards Compat.
    if not neon_skills["neon_token"]:
        try:
            neon_skills["neon_token"] = find_neon_git_token()  # TODO: GetPrivateKeys
            populate_github_token_config(neon_skills["neon_token"])
        except FileNotFoundError:
            LOG.warning(f"No Github token found; skills may fail to install!")
    skills_config = {**mycroft_config.get("skills", {}), **neon_skills}
    return skills_config


def get_neon_client_config() -> dict:
    core_config = get_neon_local_config()
    server_addr = core_config.get("remoteVars", {}).get("remoteHost", "167.172.112.7")
    if server_addr == "64.34.186.92":
        LOG.warning(f"Depreciated call to host: {server_addr}")
        server_addr = "167.172.112.7"
    return {"server_addr": server_addr,
            "devVars": core_config["devVars"],
            "remoteVars": core_config["remoteVars"]}


def get_neon_transcribe_config() -> dict:
    local_config = get_neon_local_config()
    user_config = get_neon_user_config()
    neon_transcribe_config = dict()
    neon_transcribe_config["transcript_dir"] = local_config["dirVars"].get("docsDir", "")
    neon_transcribe_config["audio_permission"] = user_config["privacy"].get("save_audio", False)
    return neon_transcribe_config


def _move_config_sections(user_config, local_config):
    """
    Temporary method to handle one-time migration of user_config params to local_config
    Args:
        user_config (NGIConfig): user configuration object
        local_config (NGIConfig): local configuration object
    """
    depreciated_user_configs = ("interface", "listener", "skills", "session", "tts", "stt", "logs", "device")
    if any([d in user_config.content for d in depreciated_user_configs]):
        LOG.warning("Depreciated keys found in user config! Adding them to local config")
        if "wake_words_enabled" in user_config.content.get("interface", dict()):
            user_config["interface"]["wake_word_enabled"] = user_config["interface"].pop("wake_words_enabled")
        config_to_move = {"interface": user_config.content.pop("interface", {}),
                          "listener": user_config.content.pop("listener", {}),
                          "skills": user_config.content.pop("skills", {}),
                          "session": user_config.content.pop("session", {}),
                          "tts": user_config.content.pop("tts", {}),
                          "stt": user_config.content.pop("stt", {}),
                          "logs": user_config.content.pop("logs", {}),
                          "device": user_config.content.pop("device", {})}
        local_config.update_keys(config_to_move)


def _safe_mycroft_config() -> dict:
    """
    Safe reference to mycroft config that always returns a dict
    Returns:
        dict mycroft configuration
    """
    try:
        mycroft = read_mycroft_config()
    except FileNotFoundError:
        mycroft = LocalConf(os.path.join(os.path.dirname(__file__), "default_configurations", "mycroft.conf"))
    return mycroft


def get_neon_user_config(path: Optional[str] = None) -> NGIConfig:
    """
    Returns a dict user configuration and handles any migration of configuration values to local config from user config
    Args:
        path: optional path to yml configuration files
    Returns:
        NGIConfig object with user config
    """
    user_config = NGIConfig("ngi_user_info", path)
    default_user_config = NGIConfig("default_user_conf",
                                    os.path.join(os.path.dirname(__file__), "default_configurations"))
    if len(user_config.content) == 0:
        LOG.info("Created Empty User Config!")
        user_config.populate(default_user_config.content)
    local_config = NGIConfig("ngi_local_conf", path)
    _move_config_sections(user_config, local_config)
    user_config.make_equal_by_keys(default_user_config.content)
    LOG.info(f"Loaded user config from {user_config.file_path}")
    return user_config


def get_neon_local_config(path: Optional[str] = None) -> NGIConfig:
    """
    Returns a dict local configuration and handles any
     migration of configuration values to local config from user config
    Args:
        path: optional path to yml configuration files
    Returns:
        NGIConfig object with local config
    """
    local_config = NGIConfig("ngi_local_conf", path)
    default_local_config = NGIConfig("default_core_conf",
                                     os.path.join(os.path.dirname(__file__), "default_configurations"))
    if len(local_config.content) == 0:
        LOG.info(f"Created Empty Local Config at {local_config.path}")
        local_config.populate(default_local_config.content)
        # TODO: Update from Mycroft config DM
    user_config = NGIConfig("ngi_user_info", path)
    _move_config_sections(user_config, local_config)
    local_config.make_equal_by_keys(default_local_config.content)
    LOG.info(f"Loaded local config from {local_config.file_path}")
    return local_config


def get_neon_device_type() -> str:
    """
    Returns device type (server, pi, other)
    Returns:
        str device type
    """
    import platform
    import importlib.util
    local_config = get_neon_local_config()
    config_dev = local_config["devVars"].get("devType", "")
    if "pi" in config_dev:
        return "pi"
    if config_dev == "server":
        return "server"
    if config_dev != "generic":
        return config_dev
    if "arm" in platform.machine():
        return "pi"
    if importlib.util.find_spec("neon-core-client"):
        return "desktop"
    if importlib.util.find_spec("neon-core-server"):
        return "server"
    return "desktop"


def is_neon_core() -> bool:
    """
    Checks for neon-specific packages to determine if this is a Neon Core or a Mycroft Core
    Returns:
        True if core is Neon, else False
    """
    import importlib.util
    if importlib.util.find_spec("neon-speech"):
        return True
    if importlib.util.find_spec("neon-core-client"):
        return True
    if importlib.util.find_spec("neon-core-server"):
        return True
    return False


def get_mycroft_compatible_config(mycroft_only=False):
    # TODO: This is kinda slow, should probably be depreciated DM
    default_config = _safe_mycroft_config()
    if mycroft_only or not is_neon_core():
        return default_config
    speech = get_neon_speech_config()
    user = get_neon_user_config()
    local = get_neon_local_config()

    default_config["lang"] = "en-us"
    default_config["language"] = get_neon_lang_config()
    default_config["keys"] = {}  # TODO: Get keys DM
    # default_config["text_parsers"]  TODO
    default_config["audio_parsers"] = speech["audio_parsers"]
    default_config["system_unit"] = user["units"]["measure"]
    default_config["time_format"] = "half" if user["units"]["time"] == 12 else "full"
    default_config["date_format"] = user["units"]["date"]
    default_config["opt_in"] = local["prefFlags"]["metrics"]
    default_config["confirm_listening"] = local["interface"]["confirm_listening"]
    default_config["sounds"] = {**default_config.get("sounds", {}), **local.get("sounds", {})}
    default_config["data_dir"] = local["dirVars"]["rootDir"]
    default_config["skills"] = get_neon_skills_config()
    default_config["server"] = get_neon_api_config()
    default_config["websocket"] = get_neon_bus_config()
    default_config["gui_websocket"] = {**default_config.get("gui_websocket", {}), **local["gui"]}
    default_config["listener"] = speech["listener"]
    # default_config["precise"]
    default_config["hotwords"] = speech["hotwords"]
    # default_config["ignore_logs"]
    default_config["session"] = local["session"]
    default_config["stt"] = speech["stt"]
    default_config["tts"] = local["tts"]
    default_config["Audio"] = get_neon_audio_config()
    # default_config["Display"]

    return default_config


def create_config_from_setup_params(path=None) -> NGIConfig:
    """
    Populate a (probably) new local config with parameters gathered during setup
    Args:
        path: Optional config path
    Returns:
        NGIConfig object generated from environment vars
    """
    local_conf = get_neon_local_config(path)
    pref_flags = local_conf["prefFlags"]
    local_conf["prefFlags"]["devMode"] = os.environ.get("devMode", str(pref_flags["devMode"])).lower() == "true"
    local_conf["prefFlags"]["autoStart"] = os.environ.get("autoStart", str(pref_flags["autoStart"])).lower() == "true"
    local_conf["prefFlags"]["autoUpdate"] = os.environ.get("autoUpdate",
                                                           str(pref_flags["autoUpdate"])).lower() == "true"
    local_conf["skills"]["neon_token"] = os.environ.get("GITHUB_TOKEN")
    local_conf["tts"]["module"] = os.environ.get("ttsModule", local_conf["tts"]["module"])
    local_conf["stt"]["module"] = os.environ.get("sttModule", local_conf["stt"]["module"])
    if os.environ.get("installServer", "false") == "true":
        local_conf["devVars"]["devType"] = "server"
    else:
        import platform
        local_conf["devVars"]["devType"] = platform.system().lower()

    local_conf["devVars"]["devName"] = os.environ.get("devName")

    if local_conf["prefFlags"]["devMode"]:
        root_path = os.environ.get("installerDir", local_conf.path)
        local_conf["dirVars"]["skillsDir"] = os.path.join(root_path, "skills")
        local_conf["dirVars"]["diagsDir"] = os.path.join(root_path, "Diagnostics")
        local_conf["dirVars"]["logsDir"] = os.path.join(root_path, "logs")
        local_conf["skills"]["default_skills"] =\
            "https://raw.githubusercontent.com/NeonGeckoCom/neon-skills-submodules/dev/.utilities/DEFAULT-SKILLS-DEV"
    else:
        local_conf["dirVars"]["logsDir"] = "~/.local/share/neon/logs"

    local_conf.write_changes()
    return local_conf



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
from typing import Optional

from mycroft_bus_client import Message


def request_from_mobile(message: Message) -> bool:
    """
    Check if a request is from a mobile device
    Args:
        message: Message object associated with request

    Returns:
        True if message is from a mobile app, else False
    """
    return message.context.get("mobile", False)


def get_message_user(message: Message) -> Optional[str]:
    """
    Get the user associated with a message
    Args:
        message: Message object associated with request

    Returns:
        Username associated with message
    """
    if not message:
        raise ValueError
    if not hasattr(message, "context"):
        raise AttributeError(type(message))
    return message.context.get("username")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
import os

import logging

from datetime import datetime, timedelta
from typing import Optional, Union

from neon_utils.logger import LOG
from neon_utils.configuration_utils import get_neon_local_config

LOG_DIR = os.path.expanduser(get_neon_local_config()["dirVars"]["logsDir"])


def remove_old_logs(log_dir: str = LOG_DIR, history_to_retain: timedelta = timedelta(weeks=6)):
    """
    Removes archived logs older than the specified history timedelta
    Args:
        log_dir: Path to archived logs
        history_to_retain: Timedelta of history to retain
    """
    from shutil import rmtree
    for archive in os.listdir(log_dir):
        archive_path = os.path.join(log_dir, archive)
        if not os.path.isdir(archive_path):
            continue
        if datetime.now() - datetime.fromtimestamp(os.path.getmtime(archive_path)) > history_to_retain:
            LOG.info(f"removing {archive}")
            rmtree(archive_path)


def archive_logs(log_dir: str = LOG_DIR, archive_dir: Optional[str] = None):
    """
    Archives the logs in the specified log_dir to log_dir/dir_name
    Args:
        log_dir: Path to log files to be archived
        archive_dir: Directory to archive logs to (defaults to formatted time)
    """
    from glob import glob
    from os.path import join, basename
    from os import makedirs
    from shutil import move
    default_dirname = "logs--" + datetime.now().strftime("%Y-%m-%d--%H:%M:%S")
    archive_dir = join(log_dir, archive_dir or default_dirname)
    makedirs(archive_dir, exist_ok=True)
    for file in glob(join(log_dir, "*.log")):
        if basename(file) != "start.log":
            move(file, archive_dir)


def get_logger(log_name: str, log_dir: str = LOG_DIR, std_out: bool = False) -> logging.Logger:
    """
    Get a logger with the specified name and write to the specified log_dir and optionally std_out
    Args:
        log_name: Name of log (also used as log filename)
        log_dir: Directory to write log file to
        std_out: Flag to include logs in STDOUT

    Returns:
        Logger with the specified handlers
    """
    LOG.init({"path": log_dir or "stdout"})
    LOG.name = log_name
    log = LOG.create_logger(log_name, std_out)
    return log


def get_log_file_for_module(module_name: Union[str, list]) -> str:
    """
    Gets the default log basename for the requested module
    Args:
        module_name: Runnable argument passed to Popen
            (i.e. neon_speech_client, [python3, -m, mycroft.skills])

    Returns:
        Path to logfile
    """
    if isinstance(module_name, list):
        module_name = module_name[-1]
    if module_name in ("neon_speech_client", "neon_speech"):
        log_name = "voice.log"
    elif module_name in ("neon_audio_client", "neon_audio"):
        log_name = "audio.log"
    elif module_name in ("neon_enclosure_client", "neon_enclosure"):
        log_name = "enclosure.log"
    elif module_name == "neon_core_client":
        log_name = "client.log"
    elif module_name == "neon_core_server":
        log_name = "server.log"
    elif module_name == "mycroft-gui-app":
        log_name = "gui.log"
    elif ".messagebus.service" in module_name:
        log_name = "bus.log"
    elif ".skills" in module_name:
        log_name = "skills.log"
    else:
        log_name = "extras.log"

    return os.path.join(LOG_DIR, log_name)




configuration_available = {'ttsVoice': {
        "en-au": {"female": "Nicole", "male": "Russell"},
        "en-gb": {"female": "Emma", "male": "Brian"},
        "en-in": {"female": "Raveena", "male": "Aditi"},
        "en-us": {"female": "Joanna", "male": "Joey"},
        "en-gb-wls": {"female": "Geraint", "male": "Geraint"}
        }
}